/*************************************************************************
Copyright (c) 2007, Sergey Bochkanov (ALGLIB project).

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses

>>> END OF LICENSE >>>
*************************************************************************/

#ifndef _matgen_h
#define _matgen_h

#include "ap.h"
#include "amp.h"
#include "reflections.h"
#include "creflections.h"
#include "hqrnd.h"
namespace matgen
{
    
    void rmatrixrndorthogonal(int n,
        ap::template_2d_array< amp::ampf >& a);
    
    void rmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::ampf >& a);
    
    void cmatrixrndorthogonal(int n,
        ap::template_2d_array< amp::campf >& a);
    
    void cmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::campf >& a);
    
    void smatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::ampf >& a);
    
    void spdmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::ampf >& a);
    
    void hmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::campf >& a);
    
    void hpdmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::campf >& a);
    
    void rmatrixrndorthogonalfromtheright(ap::template_2d_array< amp::ampf >& a,
        int m,
        int n);
    
    void rmatrixrndorthogonalfromtheleft(ap::template_2d_array< amp::ampf >& a,
        int m,
        int n);
    
    void cmatrixrndorthogonalfromtheright(ap::template_2d_array< amp::campf >& a,
        int m,
        int n);
    
    void cmatrixrndorthogonalfromtheleft(ap::template_2d_array< amp::campf >& a,
        int m,
        int n);
    
    void smatrixrndmultiply(ap::template_2d_array< amp::ampf >& a,
        int n);
    
    void hmatrixrndmultiply(ap::template_2d_array< amp::campf >& a,
        int n);


    /*************************************************************************
    Generation of a random uniformly distributed (Haar) orthogonal matrix

    INPUT PARAMETERS:
        N   -   matrix size, N>=1
        
    OUTPUT PARAMETERS:
        A   -   orthogonal NxN matrix, array[0..N-1,0..N-1]

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void rmatrixrndorthogonal(int n,
        ap::template_2d_array< amp::ampf >& a)
    {
        int i;
        int j;


        ap::ap_error::make_assertion(n>=1);
        a.setbounds(0, n-1, 0, n-1);
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                if( i==j )
                {
                    a(i,j) = 1;
                }
                else
                {
                    a(i,j) = 0;
                }
            }
        }
        rmatrixrndorthogonalfromtheright(a, n, n);
    }


    /*************************************************************************
    Generation of random NxN matrix with given condition number and norm2(A)=1

    INPUT PARAMETERS:
        N   -   matrix size
        C   -   condition number (in 2-norm)

    OUTPUT PARAMETERS:
        A   -   random matrix with norm2(A)=1 and cond(A)=C

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void rmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::ampf >& a)
    {
        int i;
        int j;
        amp::ampf l1;
        amp::ampf l2;


        ap::ap_error::make_assertion(n>=1 && c>=1);
        a.setbounds(0, n-1, 0, n-1);
        if( n==1 )
        {
            
            //
            // special case
            //
            a(0,0) = 2*ap::randominteger(2)-1;
            return;
        }
        l1 = 0;
        l2 = amp::log(1/c);
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                a(i,j) = 0;
            }
        }
        a(0,0) = amp::exp(l1);
        for(i=1; i<=n-2; i++)
        {
            a(i,i) = amp::exp(amp::ampf::getRandom()*(l2-l1)+l1);
        }
        a(n-1,n-1) = amp::exp(l2);
        rmatrixrndorthogonalfromtheleft(a, n, n);
        rmatrixrndorthogonalfromtheright(a, n, n);
    }


    /*************************************************************************
    Generation of a random Haar distributed orthogonal complex matrix

    INPUT PARAMETERS:
        N   -   matrix size, N>=1

    OUTPUT PARAMETERS:
        A   -   orthogonal NxN matrix, array[0..N-1,0..N-1]

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void cmatrixrndorthogonal(int n,
        ap::template_2d_array< amp::campf >& a)
    {
        int i;
        int j;


        ap::ap_error::make_assertion(n>=1);
        a.setbounds(0, n-1, 0, n-1);
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                if( i==j )
                {
                    a(i,j) = 1;
                }
                else
                {
                    a(i,j) = 0;
                }
            }
        }
        cmatrixrndorthogonalfromtheright(a, n, n);
    }


    /*************************************************************************
    Generation of random NxN complex matrix with given condition number C and
    norm2(A)=1

    INPUT PARAMETERS:
        N   -   matrix size
        C   -   condition number (in 2-norm)

    OUTPUT PARAMETERS:
        A   -   random matrix with norm2(A)=1 and cond(A)=C

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void cmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::campf >& a)
    {
        int i;
        int j;
        amp::ampf l1;
        amp::ampf l2;
        hqrnd::hqrndstate state;
        amp::campf v;


        ap::ap_error::make_assertion(n>=1 && c>=1);
        a.setbounds(0, n-1, 0, n-1);
        if( n==1 )
        {
            
            //
            // special case
            //
            hqrnd::hqrndrandomize(state);
            hqrnd::hqrndunit2(state, v.x, v.y);
            a(0,0) = v;
            return;
        }
        l1 = 0;
        l2 = amp::log(1/c);
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                a(i,j) = 0;
            }
        }
        a(0,0) = amp::exp(l1);
        for(i=1; i<=n-2; i++)
        {
            a(i,i) = amp::exp(amp::ampf::getRandom()*(l2-l1)+l1);
        }
        a(n-1,n-1) = amp::exp(l2);
        cmatrixrndorthogonalfromtheleft(a, n, n);
        cmatrixrndorthogonalfromtheright(a, n, n);
    }


    /*************************************************************************
    Generation of random NxN symmetric matrix with given condition number  and
    norm2(A)=1

    INPUT PARAMETERS:
        N   -   matrix size
        C   -   condition number (in 2-norm)

    OUTPUT PARAMETERS:
        A   -   random matrix with norm2(A)=1 and cond(A)=C

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void smatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::ampf >& a)
    {
        int i;
        int j;
        amp::ampf l1;
        amp::ampf l2;


        ap::ap_error::make_assertion(n>=1 && c>=1);
        a.setbounds(0, n-1, 0, n-1);
        if( n==1 )
        {
            
            //
            // special case
            //
            a(0,0) = 2*ap::randominteger(2)-1;
            return;
        }
        
        //
        // Prepare matrix
        //
        l1 = 0;
        l2 = amp::log(1/c);
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                a(i,j) = 0;
            }
        }
        a(0,0) = amp::exp(l1);
        for(i=1; i<=n-2; i++)
        {
            a(i,i) = (2*ap::randominteger(2)-1)*amp::exp(amp::ampf::getRandom()*(l2-l1)+l1);
        }
        a(n-1,n-1) = amp::exp(l2);
        
        //
        // Multiply
        //
        smatrixrndmultiply(a, n);
    }


    /*************************************************************************
    Generation of random NxN symmetric positive definite matrix with given
    condition number and norm2(A)=1

    INPUT PARAMETERS:
        N   -   matrix size
        C   -   condition number (in 2-norm)

    OUTPUT PARAMETERS:
        A   -   random SPD matrix with norm2(A)=1 and cond(A)=C

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void spdmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::ampf >& a)
    {
        int i;
        int j;
        amp::ampf l1;
        amp::ampf l2;


        
        //
        // Special cases
        //
        if( n<=0 || c<1 )
        {
            return;
        }
        a.setbounds(0, n-1, 0, n-1);
        if( n==1 )
        {
            a(0,0) = 1;
            return;
        }
        
        //
        // Prepare matrix
        //
        l1 = 0;
        l2 = amp::log(1/c);
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                a(i,j) = 0;
            }
        }
        a(0,0) = amp::exp(l1);
        for(i=1; i<=n-2; i++)
        {
            a(i,i) = amp::exp(amp::ampf::getRandom()*(l2-l1)+l1);
        }
        a(n-1,n-1) = amp::exp(l2);
        
        //
        // Multiply
        //
        smatrixrndmultiply(a, n);
    }


    /*************************************************************************
    Generation of random NxN Hermitian matrix with given condition number  and
    norm2(A)=1

    INPUT PARAMETERS:
        N   -   matrix size
        C   -   condition number (in 2-norm)

    OUTPUT PARAMETERS:
        A   -   random matrix with norm2(A)=1 and cond(A)=C

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void hmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::campf >& a)
    {
        int i;
        int j;
        amp::ampf l1;
        amp::ampf l2;


        ap::ap_error::make_assertion(n>=1 && c>=1);
        a.setbounds(0, n-1, 0, n-1);
        if( n==1 )
        {
            
            //
            // special case
            //
            a(0,0) = 2*ap::randominteger(2)-1;
            return;
        }
        
        //
        // Prepare matrix
        //
        l1 = 0;
        l2 = amp::log(1/c);
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                a(i,j) = 0;
            }
        }
        a(0,0) = amp::exp(l1);
        for(i=1; i<=n-2; i++)
        {
            a(i,i) = (2*ap::randominteger(2)-1)*amp::exp(amp::ampf::getRandom()*(l2-l1)+l1);
        }
        a(n-1,n-1) = amp::exp(l2);
        
        //
        // Multiply
        //
        hmatrixrndmultiply(a, n);
        
        //
        // post-process to ensure that matrix diagonal is real
        //
        for(i=0; i<=n-1; i++)
        {
            a(i,i).y = 0;
        }
    }


    /*************************************************************************
    Generation of random NxN Hermitian positive definite matrix with given
    condition number and norm2(A)=1

    INPUT PARAMETERS:
        N   -   matrix size
        C   -   condition number (in 2-norm)

    OUTPUT PARAMETERS:
        A   -   random HPD matrix with norm2(A)=1 and cond(A)=C

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void hpdmatrixrndcond(int n,
        amp::ampf c,
        ap::template_2d_array< amp::campf >& a)
    {
        int i;
        int j;
        amp::ampf l1;
        amp::ampf l2;


        
        //
        // Special cases
        //
        if( n<=0 || c<1 )
        {
            return;
        }
        a.setbounds(0, n-1, 0, n-1);
        if( n==1 )
        {
            a(0,0) = 1;
            return;
        }
        
        //
        // Prepare matrix
        //
        l1 = 0;
        l2 = amp::log(1/c);
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                a(i,j) = 0;
            }
        }
        a(0,0) = amp::exp(l1);
        for(i=1; i<=n-2; i++)
        {
            a(i,i) = amp::exp(amp::ampf::getRandom()*(l2-l1)+l1);
        }
        a(n-1,n-1) = amp::exp(l2);
        
        //
        // Multiply
        //
        hmatrixrndmultiply(a, n);
        
        //
        // post-process to ensure that matrix diagonal is real
        //
        for(i=0; i<=n-1; i++)
        {
            a(i,i).y = 0;
        }
    }


    /*************************************************************************
    Multiplication of MxN matrix by NxN random Haar distributed orthogonal matrix

    INPUT PARAMETERS:
        A   -   matrix, array[0..M-1, 0..N-1]
        M, N-   matrix size

    OUTPUT PARAMETERS:
        A   -   A*Q, where Q is random NxN orthogonal matrix

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void rmatrixrndorthogonalfromtheright(ap::template_2d_array< amp::ampf >& a,
        int m,
        int n)
    {
        amp::ampf tau;
        amp::ampf lambda;
        int s;
        int i;
        amp::ampf u1;
        amp::ampf u2;
        ap::template_1d_array< amp::ampf > w;
        ap::template_1d_array< amp::ampf > v;
        hqrnd::hqrndstate state;


        ap::ap_error::make_assertion(n>=1 && m>=1);
        if( n==1 )
        {
            
            //
            // Special case
            //
            tau = 2*ap::randominteger(2)-1;
            for(i=0; i<=m-1; i++)
            {
                a(i,0) = a(i,0)*tau;
            }
            return;
        }
        
        //
        // General case.
        // First pass.
        //
        w.setbounds(0, m-1);
        v.setbounds(1, n);
        hqrnd::hqrndrandomize(state);
        for(s=2; s<=n; s++)
        {
            
            //
            // Prepare random normal v
            //
            do
            {
                i = 1;
                while( i<=s )
                {
                    hqrnd::hqrndnormal2(state, u1, u2);
                    v(i) = u1;
                    if( i+1<=s )
                    {
                        v(i+1) = u2;
                    }
                    i = i+2;
                }
                lambda = amp::vdotproduct(v.getvector(1, s), v.getvector(1, s));
            }
            while( lambda==0 );
            
            //
            // Prepare and apply reflection
            //
            reflections::generatereflection(v, s, tau);
            v(1) = 1;
            reflections::applyreflectionfromtheright(a, tau, v, 0, m-1, n-s, n-1, w);
        }
        
        //
        // Second pass.
        //
        for(i=0; i<=n-1; i++)
        {
            tau = 2*ap::randominteger(2)-1;
            amp::vmul(a.getcolumn(i, 0, m-1), tau);
        }
    }


    /*************************************************************************
    Multiplication of MxN matrix by MxM random Haar distributed orthogonal matrix

    INPUT PARAMETERS:
        A   -   matrix, array[0..M-1, 0..N-1]
        M, N-   matrix size

    OUTPUT PARAMETERS:
        A   -   Q*A, where Q is random MxM orthogonal matrix

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void rmatrixrndorthogonalfromtheleft(ap::template_2d_array< amp::ampf >& a,
        int m,
        int n)
    {
        amp::ampf tau;
        amp::ampf lambda;
        int s;
        int i;
        int j;
        amp::ampf u1;
        amp::ampf u2;
        ap::template_1d_array< amp::ampf > w;
        ap::template_1d_array< amp::ampf > v;
        hqrnd::hqrndstate state;


        ap::ap_error::make_assertion(n>=1 && m>=1);
        if( m==1 )
        {
            
            //
            // special case
            //
            tau = 2*ap::randominteger(2)-1;
            for(j=0; j<=n-1; j++)
            {
                a(0,j) = a(0,j)*tau;
            }
            return;
        }
        
        //
        // General case.
        // First pass.
        //
        w.setbounds(0, n-1);
        v.setbounds(1, m);
        hqrnd::hqrndrandomize(state);
        for(s=2; s<=m; s++)
        {
            
            //
            // Prepare random normal v
            //
            do
            {
                i = 1;
                while( i<=s )
                {
                    hqrnd::hqrndnormal2(state, u1, u2);
                    v(i) = u1;
                    if( i+1<=s )
                    {
                        v(i+1) = u2;
                    }
                    i = i+2;
                }
                lambda = amp::vdotproduct(v.getvector(1, s), v.getvector(1, s));
            }
            while( lambda==0 );
            
            //
            // Prepare and apply reflection
            //
            reflections::generatereflection(v, s, tau);
            v(1) = 1;
            reflections::applyreflectionfromtheleft(a, tau, v, m-s, m-1, 0, n-1, w);
        }
        
        //
        // Second pass.
        //
        for(i=0; i<=m-1; i++)
        {
            tau = 2*ap::randominteger(2)-1;
            amp::vmul(a.getrow(i, 0, n-1), tau);
        }
    }


    /*************************************************************************
    Multiplication of MxN complex matrix by NxN random Haar distributed
    complex orthogonal matrix

    INPUT PARAMETERS:
        A   -   matrix, array[0..M-1, 0..N-1]
        M, N-   matrix size

    OUTPUT PARAMETERS:
        A   -   A*Q, where Q is random NxN orthogonal matrix

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void cmatrixrndorthogonalfromtheright(ap::template_2d_array< amp::campf >& a,
        int m,
        int n)
    {
        amp::campf lambda;
        amp::campf tau;
        int s;
        int i;
        ap::template_1d_array< amp::campf > w;
        ap::template_1d_array< amp::campf > v;
        hqrnd::hqrndstate state;
        int i_;


        ap::ap_error::make_assertion(n>=1 && m>=1);
        if( n==1 )
        {
            
            //
            // Special case
            //
            hqrnd::hqrndrandomize(state);
            hqrnd::hqrndunit2(state, tau.x, tau.y);
            for(i=0; i<=m-1; i++)
            {
                a(i,0) = a(i,0)*tau;
            }
            return;
        }
        
        //
        // General case.
        // First pass.
        //
        w.setbounds(0, m-1);
        v.setbounds(1, n);
        hqrnd::hqrndrandomize(state);
        for(s=2; s<=n; s++)
        {
            
            //
            // Prepare random normal v
            //
            do
            {
                for(i=1; i<=s; i++)
                {
                    hqrnd::hqrndnormal2(state, tau.x, tau.y);
                    v(i) = tau;
                }
                lambda = 0.0;
                for(i_=1; i_<=s;i_++)
                {
                    lambda += v(i_)*amp::conj(v(i_));
                }
            }
            while( lambda==0 );
            
            //
            // Prepare and apply reflection
            //
            creflections::complexgeneratereflection(v, s, tau);
            v(1) = 1;
            creflections::complexapplyreflectionfromtheright(a, tau, v, 0, m-1, n-s, n-1, w);
        }
        
        //
        // Second pass.
        //
        for(i=0; i<=n-1; i++)
        {
            hqrnd::hqrndunit2(state, tau.x, tau.y);
            for(i_=0; i_<=m-1;i_++)
            {
                a(i_,i) = tau*a(i_,i);
            }
        }
    }


    /*************************************************************************
    Multiplication of MxN complex matrix by MxM random Haar distributed
    complex orthogonal matrix

    INPUT PARAMETERS:
        A   -   matrix, array[0..M-1, 0..N-1]
        M, N-   matrix size

    OUTPUT PARAMETERS:
        A   -   Q*A, where Q is random MxM orthogonal matrix

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void cmatrixrndorthogonalfromtheleft(ap::template_2d_array< amp::campf >& a,
        int m,
        int n)
    {
        amp::campf tau;
        amp::campf lambda;
        int s;
        int i;
        int j;
        ap::template_1d_array< amp::campf > w;
        ap::template_1d_array< amp::campf > v;
        hqrnd::hqrndstate state;
        int i_;


        ap::ap_error::make_assertion(n>=1 && m>=1);
        if( m==1 )
        {
            
            //
            // special case
            //
            hqrnd::hqrndrandomize(state);
            hqrnd::hqrndunit2(state, tau.x, tau.y);
            for(j=0; j<=n-1; j++)
            {
                a(0,j) = a(0,j)*tau;
            }
            return;
        }
        
        //
        // General case.
        // First pass.
        //
        w.setbounds(0, n-1);
        v.setbounds(1, m);
        hqrnd::hqrndrandomize(state);
        for(s=2; s<=m; s++)
        {
            
            //
            // Prepare random normal v
            //
            do
            {
                for(i=1; i<=s; i++)
                {
                    hqrnd::hqrndnormal2(state, tau.x, tau.y);
                    v(i) = tau;
                }
                lambda = 0.0;
                for(i_=1; i_<=s;i_++)
                {
                    lambda += v(i_)*amp::conj(v(i_));
                }
            }
            while( lambda==0 );
            
            //
            // Prepare and apply reflection
            //
            creflections::complexgeneratereflection(v, s, tau);
            v(1) = 1;
            creflections::complexapplyreflectionfromtheleft(a, tau, v, m-s, m-1, 0, n-1, w);
        }
        
        //
        // Second pass.
        //
        for(i=0; i<=m-1; i++)
        {
            hqrnd::hqrndunit2(state, tau.x, tau.y);
            for(i_=0; i_<=n-1;i_++)
            {
                a(i,i_) = tau*a(i,i_);
            }
        }
    }


    /*************************************************************************
    Symmetric multiplication of NxN matrix by random Haar distributed
    orthogonal  matrix

    INPUT PARAMETERS:
        A   -   matrix, array[0..N-1, 0..N-1]
        N   -   matrix size

    OUTPUT PARAMETERS:
        A   -   Q'*A*Q, where Q is random NxN orthogonal matrix

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void smatrixrndmultiply(ap::template_2d_array< amp::ampf >& a,
        int n)
    {
        amp::ampf tau;
        amp::ampf lambda;
        int s;
        int i;
        amp::ampf u1;
        amp::ampf u2;
        ap::template_1d_array< amp::ampf > w;
        ap::template_1d_array< amp::ampf > v;
        hqrnd::hqrndstate state;


        
        //
        // General case.
        //
        w.setbounds(0, n-1);
        v.setbounds(1, n);
        hqrnd::hqrndrandomize(state);
        for(s=2; s<=n; s++)
        {
            
            //
            // Prepare random normal v
            //
            do
            {
                i = 1;
                while( i<=s )
                {
                    hqrnd::hqrndnormal2(state, u1, u2);
                    v(i) = u1;
                    if( i+1<=s )
                    {
                        v(i+1) = u2;
                    }
                    i = i+2;
                }
                lambda = amp::vdotproduct(v.getvector(1, s), v.getvector(1, s));
            }
            while( lambda==0 );
            
            //
            // Prepare and apply reflection
            //
            reflections::generatereflection(v, s, tau);
            v(1) = 1;
            reflections::applyreflectionfromtheright(a, tau, v, 0, n-1, n-s, n-1, w);
            reflections::applyreflectionfromtheleft(a, tau, v, n-s, n-1, 0, n-1, w);
        }
        
        //
        // Second pass.
        //
        for(i=0; i<=n-1; i++)
        {
            tau = 2*ap::randominteger(2)-1;
            amp::vmul(a.getcolumn(i, 0, n-1), tau);
            amp::vmul(a.getrow(i, 0, n-1), tau);
        }
    }


    /*************************************************************************
    Hermitian multiplication of NxN matrix by random Haar distributed
    complex orthogonal matrix

    INPUT PARAMETERS:
        A   -   matrix, array[0..N-1, 0..N-1]
        N   -   matrix size

    OUTPUT PARAMETERS:
        A   -   Q^H*A*Q, where Q is random NxN orthogonal matrix

      -- ALGLIB routine --
         04.12.2009
         Bochkanov Sergey
    *************************************************************************/
    
    void hmatrixrndmultiply(ap::template_2d_array< amp::campf >& a,
        int n)
    {
        amp::campf tau;
        amp::campf lambda;
        int s;
        int i;
        ap::template_1d_array< amp::campf > w;
        ap::template_1d_array< amp::campf > v;
        hqrnd::hqrndstate state;
        int i_;


        
        //
        // General case.
        //
        w.setbounds(0, n-1);
        v.setbounds(1, n);
        hqrnd::hqrndrandomize(state);
        for(s=2; s<=n; s++)
        {
            
            //
            // Prepare random normal v
            //
            do
            {
                for(i=1; i<=s; i++)
                {
                    hqrnd::hqrndnormal2(state, tau.x, tau.y);
                    v(i) = tau;
                }
                lambda = 0.0;
                for(i_=1; i_<=s;i_++)
                {
                    lambda += v(i_)*amp::conj(v(i_));
                }
            }
            while( lambda==0 );
            
            //
            // Prepare and apply reflection
            //
            creflections::complexgeneratereflection(v, s, tau);
            v(1) = 1;
            creflections::complexapplyreflectionfromtheright(a, tau, v, 0, n-1, n-s, n-1, w);
            creflections::complexapplyreflectionfromtheleft(a, amp::conj(tau), v, n-s, n-1, 0, n-1, w);
        }
        
        //
        // Second pass.
        //
        for(i=0; i<=n-1; i++)
        {
            hqrnd::hqrndunit2(state, tau.x, tau.y);
            for(i_=0; i_<=n-1;i_++)
            {
                a(i_,i) = tau*a(i_,i);
            }
            tau = amp::conj(tau);
            for(i_=0; i_<=n-1;i_++)
            {
                a(i,i_) = tau*a(i,i_);
            }
        }
    }
} // namespace

#endif
