//**********************************************************************************
// ExtMEM auxiliary definitions and implementations
//
// This file contains the definitions for an arbitrary precision Complex,
// Vector and Matrix datatype called HPComplex, gslgmp_vector, gslgmp_matrix
// respectively. They are derived from the corresponding double precision classes 
// provided by the GNU scientific library. All of the functions given in this
// file that operate on the above mentioned vector and matrix datatypes are also 
// derived from the corresponding GNU GSL routines.
//
// Copyright (C) 2011  Dr. Alexander Rothkopf
// 
//**********************************************************************************

//**********************************************************************************
//
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 2 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with this program; if not, write to the Free Software Foundation, Inc.,
//    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//    
//    A copy of the GNU General Public License is also available at
//    http://www.fsf.org/licensing/licenses
//
//**********************************************************************************

//**********************************************************************************
//  GMP DATATYPES DEFINITION AND BASIC INITIALIZATION AND DEALLOCATION ROUTINES
//**********************************************************************************

#define GSL_MIN(a,b) ((a) < (b) ? (a) : (b))
#define OFFSET(N, incX) ((incX) > 0 ?  0 : ((N) - 1) * (-(incX)))

#define gmpFORM %1.3e

//**********************************************************************************
//  HIGH PRECISION COMPLEX DATATYPE
//**********************************************************************************

class HPComplex {
  mpfr_t re;
  mpfr_t im;

  public:
  HPComplex();
  HPComplex(const HPComplex& a);
  HPComplex(const double& a);
  ~HPComplex();

  HPComplex& operator=(const HPComplex& rhs);
  HPComplex& operator=(const double& rhs);

  HPComplex& operator+=(const HPComplex& rhs);
  HPComplex& operator-=(const HPComplex& rhs);

  HPComplex operator*(const HPComplex& rhs);
  HPComplex operator+(const HPComplex& rhs);
  HPComplex operator-(const HPComplex& rhs);
  HPComplex operator*(const double& rhs);

  HPComplex operator/(const double& rhs);
  HPComplex& operator/=(const double& rhs);
  HPComplex& operator/=(const int& rhs);

  void addreal(const mpfr_t a);
  void addimag(const mpfr_t a);

  void addreal(double a);
  void addimag(double a);

  void real(const mpfr_t a);
  void imag(const mpfr_t a);
  void real(const double& a);
  void imag(const double& a);
  void real(const int a);
  void imag(const int a);

  double real() const;
  double imag() const;


  void Abs(mpfr_t a);
  void GetReal(mpfr_t a);
  void GetImag(mpfr_t a);



};

void HPComplex::GetReal(mpfr_t a){

	mpfr_set(a,re,GMP_RNDN);

}

void HPComplex::GetImag(mpfr_t a){

	mpfr_set(a,im,GMP_RNDN);

}



double HPComplex::real() const
{

	return mpfr_get_d(re,GMP_RNDN);

}

double HPComplex::imag() const
{

	return mpfr_get_d(im,GMP_RNDN);

}

void HPComplex::Abs(mpfr_t a){

	mpfr_t tmp1,tmp2;
	mpfr_init2(tmp1,PRECBITS);
	mpfr_init2(tmp2,PRECBITS);

	mpfr_pow_ui(tmp1,re,2,GMP_RNDN);
	mpfr_pow_ui(tmp2,im,2,GMP_RNDN);

	mpfr_add(a,tmp1,tmp2,GMP_RNDN);

	mpfr_sqrt(a,a,GMP_RNDN);	

	mpfr_clear(tmp1);
	mpfr_clear(tmp2);

}


void HPComplex::addreal(const mpfr_t a){

	mpfr_add(re,re,a,GMP_RNDN);

}

void HPComplex::addimag(const mpfr_t a){

	mpfr_add(im,im,a,GMP_RNDN);

}

void HPComplex::addreal(double a){

	mpfr_add_d(re,re,a,GMP_RNDN);

}

void HPComplex::addimag(double a){

	mpfr_add_d(im,im,a,GMP_RNDN);

}


void HPComplex::real(const mpfr_t a){

	mpfr_set(re,a,GMP_RNDN);

}

void HPComplex::imag(const mpfr_t a){

	mpfr_set(im,a,GMP_RNDN);

}

void HPComplex::real(const double& a){

	mpfr_set_d(re,a,GMP_RNDN);

}

void HPComplex::imag(const double& a){

	mpfr_set_d(im,a,GMP_RNDN);

}

void HPComplex::real(const int a){

	mpfr_set_ui(re,a,GMP_RNDN);

}

void HPComplex::imag(const int a){

	mpfr_set_ui(im,a,GMP_RNDN);

}


HPComplex& HPComplex::operator=(const HPComplex& rhs) {

	if (this!=&rhs) 
	{
		mpfr_set(re,rhs.re,GMP_RNDN);
		mpfr_set(im,rhs.im,GMP_RNDN);
	}

	return *this;

}

HPComplex& HPComplex::operator=(const double& rhs) {

	mpfr_set_d(re,rhs,GMP_RNDN);
	mpfr_set_d(im,0.,GMP_RNDN);
	
	return *this;
}

HPComplex& HPComplex::operator+=(const HPComplex& rhs) {

	mpfr_add(this->re,this->re,rhs.re,GMP_RNDN);
	mpfr_add(this->im,this->im,rhs.im,GMP_RNDN);

	return *this;
}

HPComplex& HPComplex::operator-=(const HPComplex& rhs) {

	mpfr_sub(this->re,this->re,rhs.re,GMP_RNDN);
	mpfr_sub(this->im,this->im,rhs.im,GMP_RNDN);

	return *this;
}


HPComplex HPComplex::operator*(const HPComplex& rhs) {

	HPComplex TMP;

	mpfr_t tmp1;
	mpfr_init2(tmp1,PRECBITS);

	mpfr_mul(TMP.re,this->re,rhs.re,GMP_RNDN);
	mpfr_mul(tmp1,this->im,rhs.im,GMP_RNDN);
	mpfr_sub(TMP.re,TMP.re,tmp1,GMP_RNDN);

	mpfr_mul(TMP.im,this->re,rhs.im,GMP_RNDN);
	mpfr_mul(tmp1,this->im,rhs.re,GMP_RNDN);
	mpfr_add(TMP.im,TMP.im,tmp1,GMP_RNDN);

	mpfr_clear(tmp1);

	return TMP;
}

HPComplex HPComplex::operator+(const HPComplex& rhs) {

	HPComplex TMP;

	mpfr_add(TMP.re,this->re,rhs.re,GMP_RNDN);
	mpfr_add(TMP.im,this->im,rhs.im,GMP_RNDN);

	return TMP;
}

HPComplex HPComplex::operator-(const HPComplex& rhs) {

	HPComplex TMP;

	mpfr_sub(TMP.re,this->re,rhs.re,GMP_RNDN);
	mpfr_sub(TMP.im,this->im,rhs.im,GMP_RNDN);

	return TMP;
}


HPComplex HPComplex::operator*(const double& rhs) {

	HPComplex TMP;

	mpfr_mul_d(TMP.re,this->re,rhs,GMP_RNDN);
	mpfr_mul_d(TMP.im,this->im,rhs,GMP_RNDN);

	return TMP;
}


HPComplex HPComplex::operator/(const double& rhs) {

	HPComplex TMP;

	mpfr_div_d(TMP.re,this->re,rhs,GMP_RNDN);
	mpfr_div_d(TMP.im,this->im,rhs,GMP_RNDN);

	return TMP;
}


HPComplex& HPComplex::operator/=(const double& rhs) {

	mpfr_div_d(this->re,this->re,rhs,GMP_RNDN);
	mpfr_div_d(this->im,this->im,rhs,GMP_RNDN);

	return *this;
}

HPComplex& HPComplex::operator/=(const int& rhs) {

	mpfr_div_si(this->re,this->re,rhs,GMP_RNDN);
	mpfr_div_si(this->im,this->im,rhs,GMP_RNDN);

	return *this;
}


HPComplex::HPComplex() {

  mpfr_init2(re,PRECBITS);
  mpfr_init2(im,PRECBITS);

  mpfr_set_ui(re,0,GMP_RNDN);
  mpfr_set_ui(im,0,GMP_RNDN);


}

HPComplex::HPComplex(const HPComplex& a) {

  mpfr_init2(re,PRECBITS);
  mpfr_init2(im,PRECBITS);

  mpfr_set(re,a.re,GMP_RNDN);
  mpfr_set(im,a.im,GMP_RNDN);

}

HPComplex::HPComplex(const double& a) {

  mpfr_init2(re,PRECBITS);
  mpfr_init2(im,PRECBITS);

  mpfr_set_d(re,a,GMP_RNDN);

}


HPComplex::~HPComplex() {

  mpfr_clear(re);
  mpfr_clear(im);

}

//**********************************************************************************
//  HIGH PRECISION REAL MATRICES OF NxM SIZE
//**********************************************************************************


class gslgmp_matrix {

	public:

	int size1;
	int size2;
	int memstep;
	mpfr_t * data;

};


void gslgmp_matrix_free(gslgmp_matrix* in )
{
	for(int i=0;i<in->size1*in->size2;i++){
	mpfr_clear(in->data[i]);
	}

	free(in->data);

	free(in);

}

gslgmp_matrix* gslgmp_matrix_calloc(int NUM1, int NUM2 )
{
	gslgmp_matrix* tmp;
	tmp=(gslgmp_matrix*)malloc(sizeof(gslgmp_matrix));

	
	tmp->data=(mpfr_t*)malloc(sizeof(mpfr_t)*NUM1*NUM2);

	for(int i=0;i<NUM1*NUM2;i++){
	mpfr_init2(tmp->data[i],PRECBITS);
	mpfr_set_d(tmp->data[i],0.,GMP_RNDN);
	}

	tmp->size1=NUM1;	
	tmp->size2=NUM2;
	tmp->memstep=NUM2;
	
	return tmp;
}



//**********************************************************************************
//  GMP MATRIX ACCESSING AND BASIC MANIPULATION ROUTINES + SIMPLE PRINT ROUTINE
//**********************************************************************************

void gslgmp_matrix_set(gslgmp_matrix* in, int pos1, int pos2, mpfr_t value)
{
	//Memeory Layout: for MxN   (i,j)	i*N+j

	if(pos1>=in->size1||pos2>=in->size2) {fprintf(stderr, "Counter exceeds matrix dimension, exiting!\n"); exit(-1);}
	mpfr_set(in->data[pos1*(in->memstep)+pos2],value,GMP_RNDN);
}

void gslgmp_matrix_set_ui(gslgmp_matrix* in, int pos1, int pos2, unsigned int value)
{
	//Memeory Layout: for MxN   (i,j)	i*N+j

	if(pos1>=in->size1||pos2>=in->size2) {fprintf(stderr, "Counter exceeds matrix dimension, exiting!\n"); exit(-1);}
	mpfr_set_ui(in->data[pos1*(in->memstep)+pos2],value,GMP_RNDN);
}

void gslgmp_matrix_scale(gslgmp_matrix* in, mpfr_t value)
{
	//Memeory Layout: for MxN   (i,j)	i*N+j
	
	mpfr_t tmp;
	mpfr_init2(tmp,PRECBITS);

	for(int i=0;i<in->size1;i++){
	  for(int j=0;j<in->size2;j++){
	mpfr_mul(tmp,in->data[i*(in->memstep)+j],value,GMP_RNDN);
	mpfr_set(in->data[i*(in->memstep)+j],tmp,GMP_RNDN);
	}}

	mpfr_clear(tmp);
}

void gslgmp_matrix_scale_d(gslgmp_matrix* in, double value)
{
	//Memeory Layout: for MxN   (i,j)	i*N+j
	
	mpfr_t tmp;
	mpfr_init2(tmp,PRECBITS);
	mpfr_t tmp2;
	mpfr_init2(tmp2,PRECBITS);

	mpfr_set_d(tmp2,value,GMP_RNDN);

	for(int i=0;i<in->size1;i++){
	  for(int j=0;j<in->size2;j++){
	mpfr_mul(tmp,in->data[i*(in->memstep)+j],tmp2,GMP_RNDN);
	mpfr_set(in->data[i*(in->memstep)+j],tmp,GMP_RNDN);
	}}

	mpfr_clear(tmp);
	mpfr_clear(tmp2);
}
void gslgmp_matrix_set_d(gslgmp_matrix* in, int pos1, int pos2, double value)
{
	//Memeory Layout: for MxN   (i,j)	i*N+j

	if(pos1>=in->size1||pos2>=in->size2) {fprintf(stderr, "Counter exceeds matrix dimension, exiting!\n"); exit(-1);}
	mpfr_set_d(in->data[pos1*(in->memstep)+pos2],value,GMP_RNDN);

}

void gslgmp_matrix_get(gslgmp_matrix* in, int pos1, int pos2,mpfr_t out)
{
	if(pos1>=in->size1||pos2>=in->size2) {fprintf(stderr, "Counter exceeds matrix dimension, exiting!\n"); exit(-1);}
	mpfr_set(out,in->data[pos1*(in->memstep)+pos2],GMP_RNDN);

}

void gslgmp_matrix_get_add(gslgmp_matrix* in, int pos1, int pos2,mpfr_t out)
{

	if(pos1>=in->size1||pos2>=in->size2) {fprintf(stderr, "Counter exceeds matrix dimension, exiting!\n"); exit(-1);}
	mpfr_t tmp;
	mpfr_init2(tmp,PRECBITS);

	mpfr_set(tmp,out,GMP_RNDN);
	mpfr_add(out,in->data[pos1*(in->memstep)+pos2],tmp,GMP_RNDN);

	mpfr_clear(tmp);
}

double gslgmp_matrix_get_d(gslgmp_matrix* in, int pos1, int pos2)
{
	if(pos1>=in->size1||pos2>=in->size2) {fprintf(stderr, "Counter exceeds matrix dimension, exiting!\n"); exit(-1);}
	return mpfr_get_d(in->data[pos1*(in->memstep)+pos2],GMP_RNDN);

}

void gslgmp_matrix_set_identity(gslgmp_matrix* in)
{
	for(int i=0;i<in->size1;i++){
	   for(int j=0;j<in->size2;j++){
	((i==j)?mpfr_set_ui(in->data[i*(in->memstep)+j],1,GMP_RNDN):mpfr_set_ui(in->data[i*(in->memstep)+j],0,GMP_RNDN));
	}}
}


void gmpPMat(gslgmp_matrix* in)
{
	fprintf(stderr,"\n\n{");
	for(int i=0;i<in->size1;i++){
	  fprintf(stderr,"{");
	  for(int j=0;j<in->size2;j++){
	    if(j==(in->size2-1)){fprintf(stderr,"%.4e\t",gslgmp_matrix_get_d(in,i,j));}else{fprintf(stderr,"%.4e,\t",gslgmp_matrix_get_d(in,i,j));}
	  } fprintf(stderr,"},\n");
	}	fprintf(stderr,"}\n\n");

}

void gmpPMat_file(gslgmp_matrix* in, char* name)
{
	FILE* tmp;
	char fname[300];
	sprintf(fname,"gmpPMat-%s.dat",name);

	tmp=fopen(fname,"w");

	fprintf(tmp,"{");
	for(int i=0;i<in->size1;i++){
	  fprintf(tmp,"{");
	  for(int j=0;j<in->size2;j++){
	    if(j==(in->size2-1)){fprintf(tmp,"%.8e\t",gslgmp_matrix_get_d(in,i,j));}else{fprintf(tmp,"%.8e,\t",gslgmp_matrix_get_d(in,i,j));}
	  } fprintf(tmp,"},\n");
	}	fprintf(tmp,"}\n\n");

	fclose(tmp);

}

void gmpPMatDiag(gslgmp_matrix* in)
{
	if(in->size1!=in->size2){fprintf(stderr, "Diagonal Elements only for square matrix!\n");exit(-1);}
	fprintf(stderr,"\n\n{");
	  for(int j=0;j<in->size1;j++){
	    fprintf(stderr,"%.4e,\t",gslgmp_matrix_get_d(in,j,j));
	  } fprintf(stderr,"}\n\n");

}

//**********************************************************************************
//  GMP VECTOR DEFINITION AND BASIC INITIALIZATION AND DEALLOCATION ROUTINES
//**********************************************************************************

typedef struct
     {
       int size;
       mpfr_t * data;
     } gslgmp_vector;



void gslgmp_vector_free(gslgmp_vector* in )
{
	for(int i=0;i<in->size;i++){
	mpfr_clear(in->data[i]);
	}

	free(in->data);

	free(in);

}

gslgmp_vector* gslgmp_vector_calloc(int NUM1)
{
	gslgmp_vector* tmp;
	tmp=(gslgmp_vector*)malloc(sizeof(gslgmp_vector));

	
	tmp->data=(mpfr_t*)malloc(sizeof(mpfr_t)*NUM1);

	for(int i=0;i<NUM1;i++){
	mpfr_init2(tmp->data[i],PRECBITS);
	mpfr_set_d(tmp->data[i],0.,GMP_RNDN);
	}

	tmp->size=NUM1;	

	return tmp;
	
	gslgmp_vector_free(tmp);

}


//**********************************************************************************
//  GMP VECTOR ACCESSING AND BASIC MANIPULATION ROUTINES + SIMPLE PRINT ROUTINE
//**********************************************************************************

void gslgmp_vector_set(gslgmp_vector* in, int pos1, mpfr_t value)
{
	if(pos1>=in->size) {fprintf(stderr, "Counter exceeds vector dimension, exiting!\n"); exit(-1);}
	mpfr_set(in->data[pos1],value,GMP_RNDN);
}

void gslgmp_vector_set_d(gslgmp_vector* in, int pos1, double value)
{
	if(pos1>=in->size) {fprintf(stderr, "Counter exceeds vector dimension, exiting!\n"); exit(-1);}
	mpfr_set_d(in->data[pos1],value,GMP_RNDN);

}

void gslgmp_vector_set_ui(gslgmp_vector* in, int pos1, unsigned int value)
{
	if(pos1>=in->size) {fprintf(stderr, "Counter exceeds vector dimension, exiting!\n"); exit(-1);}
	mpfr_set_ui(in->data[pos1],value,GMP_RNDN);

}

void gslgmp_vector_set_zero(gslgmp_vector* in )
{
	for(int i=0;i<in->size;i++){
	mpfr_set_d(in->data[i],0.,GMP_RNDN);
	}

}

void gslgmp_vector_get(gslgmp_vector* in, int pos1,mpfr_t out)
{
	if(pos1>=in->size) {fprintf(stderr, "Counter exceeds vector dimension, exiting!\n"); exit(-1);}
	mpfr_set(out,in->data[pos1],GMP_RNDN);

}

void gslgmp_vector_get_add(gslgmp_vector* in, int pos1,mpfr_t out)
{
	if(pos1>=in->size) {fprintf(stderr, "Counter exceeds vector dimension, exiting!\n"); exit(-1);}
	mpfr_t tmp;
	mpfr_init2(tmp,PRECBITS);

	mpfr_set(tmp,out,GMP_RNDN);
	mpfr_add(out,in->data[pos1],tmp,GMP_RNDN);

	mpfr_clear(tmp);
}


double gslgmp_vector_get_d(gslgmp_vector* in, int pos1)
{
	if(pos1>=in->size) {fprintf(stderr, "Counter exceeds vector dimension, exiting!\n"); exit(-1);}
	return mpfr_get_d(in->data[pos1],GMP_RNDN);

}


void gmpPVec(gslgmp_vector* in)
{
	fprintf(stderr,"\n\n{");
	  for(int j=0;j<in->size;j++){
	    fprintf(stderr,"%.4e,\t",gslgmp_vector_get_d(in,j));
	  } fprintf(stderr,"}\n");
}

//**********************************************************************************
//  GMP VECTOR AND MATRIX MULTIPLICATION ROUTINES
//**********************************************************************************

void gmp_mult2m(gslgmp_matrix* in1, int t1,gslgmp_matrix* in2, int t2, gslgmp_matrix* out)
{

  mpfr_t tmp1;
  mpfr_init2(tmp1,PRECBITS);
  mpfr_t tmp2;
  mpfr_init2(tmp2,PRECBITS);


  if(t1==0 && t2==0){
  	if(in1->size2!=in2->size1||out->size1!=in1->size1||out->size2!=in2->size2) {fprintf(stderr,"Matrix mult Dimsension do not match, exiting\n");exit(-1);}

	for(int i=0;i<in1->size1;i++){
	  for(int j=0;j<in2->size2;j++){
		mpfr_set_d(tmp1,0.0,GMP_RNDN);
		for(int k=0;k<in1->size2;k++){
		mpfr_mul(tmp2,in1->data[i*(in1->memstep)+k],in2->data[k*(in2->memstep)+j],GMP_RNDN);
		mpfr_add(tmp1,tmp2,tmp1,GMP_RNDN);
		}

		mpfr_set(out->data[i*(out->memstep)+j],tmp1,GMP_RNDN);

	}}

  }

  if(t1==1 && t2==0){
  	if(in1->size1!=in2->size1||out->size1!=in1->size2||out->size2!=in2->size2) {fprintf(stderr,"Matrix mult Dimsension do not match, exiting\n");exit(-1);}

	for(int i=0;i<in1->size2;i++){
	  for(int j=0;j<in2->size2;j++){
		mpfr_set_d(tmp1,0.0,GMP_RNDN);
		for(int k=0;k<in1->size1;k++){
		mpfr_mul(tmp2,in1->data[k*(in1->memstep)+i],in2->data[k*(in2->memstep)+j],GMP_RNDN);
		mpfr_add(tmp1,tmp2,tmp1,GMP_RNDN);
		}

		mpfr_set(out->data[i*(out->memstep)+j],tmp1,GMP_RNDN);

	}}



  }

  if(t1==0 && t2==1){
  	if(in1->size2!=in2->size2||out->size1!=in1->size1||out->size2!=in2->size1) {fprintf(stderr,"Matrix mult Dimsension do not match, exiting\n");exit(-1);}

	for(int i=0;i<in1->size1;i++){
	  for(int j=0;j<in2->size1;j++){
		mpfr_set_d(tmp1,0.0,GMP_RNDN);
		for(int k=0;k<in1->size2;k++){
		mpfr_mul(tmp2,in1->data[i*(in1->memstep)+k],in2->data[j*(in2->memstep)+k],GMP_RNDN);
		mpfr_add(tmp1,tmp2,tmp1,GMP_RNDN);
		}

		mpfr_set(out->data[i*(out->memstep)+j],tmp1,GMP_RNDN);

	}}


  }

  if(t1==1 && t2==1){
  	if(in1->size1!=in2->size2||out->size1!=in1->size2||out->size2!=in2->size1) {fprintf(stderr,"Matrix mult Dimsension do not match, exiting\n");exit(-1);}

	for(int i=0;i<in1->size2;i++){
	  for(int j=0;j<in2->size1;j++){
		mpfr_set_d(tmp1,0.0,GMP_RNDN);
		for(int k=0;k<in1->size1;k++){
		mpfr_mul(tmp2,in1->data[k*(in1->memstep)+i],in2->data[j*(in2->memstep)+k],GMP_RNDN);
		mpfr_add(tmp1,tmp2,tmp1,GMP_RNDN);
		}

		mpfr_set(out->data[i*(out->memstep)+j],tmp1,GMP_RNDN);

	}}


  }

	mpfr_clear(tmp1);
	mpfr_clear(tmp2);

}

void gmp_mult2m_conv(gslgmp_matrix* in1, int t1,gslgmp_matrix* in2, int t2, gslgmp_matrix* out)
{

  mpfr_t tmp1;
  mpfr_init2(tmp1,PRECBITS);
  mpfr_t tmp2;
  mpfr_init2(tmp2,PRECBITS);
  mpfr_t tmp3;
  mpfr_init2(tmp3,PRECBITS);
  mpfr_t hlf;
  mpfr_init2(hlf,PRECBITS);
  mpfr_set_d(hlf,0.5,GMP_RNDN);

  if(t1==0 && t2==0){
  	if(in1->size2!=in2->size1||out->size1!=in1->size1||out->size2!=in2->size2) {fprintf(stderr,"Matrix mult Dimsension do not match, exiting\n");exit(-1);}

	for(int i=0;i<in1->size1;i++){
	  for(int j=0;j<in2->size2;j++){
		mpfr_set_d(tmp1,0.0,GMP_RNDN);

		//first step in integral 1/2
		mpfr_mul(tmp2,in1->data[i*(in1->memstep)+0],in2->data[0*(in2->memstep)+j],GMP_RNDN);
		mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
		mpfr_add(tmp1,tmp3,tmp1,GMP_RNDN);


		for(int k=1;k<in1->size2-1;k++){
		mpfr_mul(tmp2,in1->data[i*(in1->memstep)+k],in2->data[k*(in2->memstep)+j],GMP_RNDN);
		mpfr_add(tmp1,tmp2,tmp1,GMP_RNDN);
		}

		//last step in integral 1/2
		mpfr_mul(tmp2,in1->data[i*(in1->memstep)+(in1->size2-1)],in2->data[(in1->size2-1)*(in2->memstep)+j],GMP_RNDN);
		mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
		mpfr_add(tmp1,tmp3,tmp1,GMP_RNDN);

		mpfr_set(out->data[i*(out->memstep)+j],tmp1,GMP_RNDN);

	}}

  }

  if(t1==1 && t2==0){
  	if(in1->size1!=in2->size1||out->size1!=in1->size2||out->size2!=in2->size2) {fprintf(stderr,"Matrix mult Dimsension do not match, exiting\n");exit(-1);}

	for(int i=0;i<in1->size2;i++){
	  for(int j=0;j<in2->size2;j++){
		mpfr_set_d(tmp1,0.0,GMP_RNDN);

		//first step 1/2
		mpfr_mul(tmp2,in1->data[0*(in1->memstep)+i],in2->data[0*(in2->memstep)+j],GMP_RNDN);
		mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
		mpfr_add(tmp1,tmp3,tmp1,GMP_RNDN);


		for(int k=1;k<in1->size1-1;k++){
		mpfr_mul(tmp2,in1->data[k*(in1->memstep)+i],in2->data[k*(in2->memstep)+j],GMP_RNDN);
		mpfr_add(tmp1,tmp2,tmp1,GMP_RNDN);
		}

		//last step 1/2
		mpfr_mul(tmp2,in1->data[(in1->size1-1)*(in1->memstep)+i],in2->data[(in1->size1-1)*(in2->memstep)+j],GMP_RNDN);
		mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
		mpfr_add(tmp1,tmp3,tmp1,GMP_RNDN);



		mpfr_set(out->data[i*(out->memstep)+j],tmp1,GMP_RNDN);

	}}



  }

  if(t1==0 && t2==1){
  	if(in1->size2!=in2->size2||out->size1!=in1->size1||out->size2!=in2->size1) {fprintf(stderr,"Matrix mult Dimsension do not match, exiting\n");exit(-1);}

	for(int i=0;i<in1->size1;i++){
	  for(int j=0;j<in2->size1;j++){
		mpfr_set_d(tmp1,0.0,GMP_RNDN);

		//First step 1/2
		mpfr_mul(tmp2,in1->data[i*(in1->memstep)+0],in2->data[j*(in2->memstep)+0],GMP_RNDN);
		mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
		mpfr_add(tmp1,tmp3,tmp1,GMP_RNDN);	


		for(int k=1;k<in1->size2-1;k++){
		mpfr_mul(tmp2,in1->data[i*(in1->memstep)+k],in2->data[j*(in2->memstep)+k],GMP_RNDN);
		mpfr_add(tmp1,tmp2,tmp1,GMP_RNDN);
		}

		//Last step 1/2
		mpfr_mul(tmp2,in1->data[i*(in1->memstep)+(in1->size2-1)],in2->data[j*(in2->memstep)+(in1->size2-1)],GMP_RNDN);
		mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
		mpfr_add(tmp1,tmp3,tmp1,GMP_RNDN);	



		mpfr_set(out->data[i*(out->memstep)+j],tmp1,GMP_RNDN);

	}}


  }

  if(t1==1 && t2==1){
  	if(in1->size1!=in2->size2||out->size1!=in1->size2||out->size2!=in2->size1) {fprintf(stderr,"Matrix mult Dimsension do not match, exiting\n");exit(-1);}

	for(int i=0;i<in1->size2;i++){
	  for(int j=0;j<in2->size1;j++){
		mpfr_set_d(tmp1,0.0,GMP_RNDN);

		//First step 1/2
		mpfr_mul(tmp2,in1->data[0*(in1->memstep)+i],in2->data[j*(in2->memstep)+0],GMP_RNDN);
		mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
		mpfr_add(tmp1,tmp3,tmp1,GMP_RNDN);	


		for(int k=1;k<in1->size1-1;k++){
		mpfr_mul(tmp2,in1->data[k*(in1->memstep)+i],in2->data[j*(in2->memstep)+k],GMP_RNDN);
		mpfr_add(tmp1,tmp2,tmp1,GMP_RNDN);
		}

		//First step 1/2
		mpfr_mul(tmp2,in1->data[(in1->size1-1)*(in1->memstep)+i],in2->data[j*(in2->memstep)+(in1->size1-1)],GMP_RNDN);
		mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
		mpfr_add(tmp1,tmp3,tmp1,GMP_RNDN);	


		mpfr_set(out->data[i*(out->memstep)+j],tmp1,GMP_RNDN);

	}}


  }

	mpfr_clear(tmp1);
	mpfr_clear(tmp2);
	mpfr_clear(tmp3);
	mpfr_clear(hlf);

}

void gmp_multmv(gslgmp_matrix* inm,int t,gslgmp_vector* inv,gslgmp_vector* out)
{
 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_t tmp2;
 mpfr_init2(tmp2,PRECBITS);

 
 if(t==0){
   if(inm->size1!=out->size){fprintf(stderr,"vector incompatible\n!");exit(-1);}


   if(inv->size!=inm->size2){fprintf(stderr,"matrix and vector incompatible\n!");exit(-1);}

   for(int i=0;i<inm->size1;i++){
     mpfr_set_d(tmp,0.0,GMP_RNDN);
     	for(int j=0;j<inm->size2;j++){
	mpfr_mul(tmp2,inm->data[i*(inm->memstep)+j],inv->data[j],GMP_RNDN);
	mpfr_add(tmp,tmp2,tmp,GMP_RNDN);
	}
	mpfr_set(out->data[i],tmp,GMP_RNDN);
   }

 }else{

    if(inm->size2!=out->size){fprintf(stderr,"vector incompatible\n!");exit(-1);}

   if(inv->size!=inm->size1){fprintf(stderr,"matrix and vector incompatible\n!");exit(-1);}

   for(int i=0;i<inm->size2;i++){
     mpfr_set_d(tmp,0.0,GMP_RNDN);
     	for(int j=0;j<inm->size1;j++){
	mpfr_mul(tmp2,inm->data[j*(inm->memstep)+i],inv->data[j],GMP_RNDN);
	mpfr_add(tmp,tmp2,tmp,GMP_RNDN);
	}
	mpfr_set(out->data[i],tmp,GMP_RNDN);
 
   }
 }

	mpfr_clear(tmp);
	mpfr_clear(tmp2);

}

void gmp_multmv_lower(gslgmp_matrix* inm,gslgmp_vector* inv,gslgmp_vector* out)
{
 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_t tmp2;
 mpfr_init2(tmp2,PRECBITS);

 
   if(inm->size1!=out->size){fprintf(stderr,"vector incompatible\n!");exit(-1);}


   if(inv->size!=inm->size2){fprintf(stderr,"matrix and vector incompatible\n!");exit(-1);}

   for(int i=0;i<inm->size1;i++){
     mpfr_set_d(tmp,0.0,GMP_RNDN);
     	for(int j=0;j<=i;j++){
	mpfr_mul(tmp2,inm->data[i*(inm->memstep)+j],inv->data[j],GMP_RNDN);
	mpfr_add(tmp,tmp2,tmp,GMP_RNDN);
	}
    	for(int j=i+1;j<inm->size2;j++){
	mpfr_mul(tmp2,inm->data[j*(inm->memstep)+i],inv->data[j],GMP_RNDN);
	mpfr_add(tmp,tmp2,tmp,GMP_RNDN);
	}


	mpfr_set(out->data[i],tmp,GMP_RNDN);
   }

	mpfr_clear(tmp);
	mpfr_clear(tmp2);

}


void gmp_multmv_conv(gslgmp_matrix* inm,int t,gslgmp_vector* inv,gslgmp_vector* out)
{
 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_t tmp2;
 mpfr_init2(tmp2,PRECBITS);
 mpfr_t tmp3;
 mpfr_init2(tmp3,PRECBITS);
 mpfr_t hlf;
 mpfr_init2(hlf,PRECBITS);
 mpfr_set_d(hlf,0.5,GMP_RNDN);
 
 if(t==0){
   if(inm->size1!=out->size){fprintf(stderr,"vector incompatible\n!");exit(-1);}


   if(inv->size!=inm->size2){fprintf(stderr,"matrix and vector incompatible\n!");exit(-1);}

   for(int i=0;i<inm->size1;i++){
     mpfr_set_d(tmp,0.0,GMP_RNDN);

	//First step 1/2
	mpfr_mul(tmp2,inm->data[i*(inm->memstep)+0],inv->data[0],GMP_RNDN);
	mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
	mpfr_add(tmp,tmp3,tmp,GMP_RNDN);

     	for(int j=1;j<inm->size2-1;j++){
	mpfr_mul(tmp2,inm->data[i*(inm->memstep)+j],inv->data[j],GMP_RNDN);
	mpfr_add(tmp,tmp2,tmp,GMP_RNDN);
	}

	//Last step 1/2
	mpfr_mul(tmp2,inm->data[i*(inm->memstep)+(inm->size2-1)],inv->data[(inm->size2-1)],GMP_RNDN);
	mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
	mpfr_add(tmp,tmp3,tmp,GMP_RNDN);


	mpfr_set(out->data[i],tmp,GMP_RNDN);
   }

 }else{

    if(inm->size2!=out->size){fprintf(stderr,"vector incompatible\n!");exit(-1);}

   if(inv->size!=inm->size1){fprintf(stderr,"matrix and vector incompatible\n!");exit(-1);}

   for(int i=0;i<inm->size2;i++){
     mpfr_set_d(tmp,0.0,GMP_RNDN);

	//First step 1/2
	mpfr_mul(tmp2,inm->data[0*(inm->memstep)+i],inv->data[0],GMP_RNDN);
	mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
	mpfr_add(tmp,tmp3,tmp,GMP_RNDN);


     	for(int j=1;j<inm->size1-1;j++){
	mpfr_mul(tmp2,inm->data[j*(inm->memstep)+i],inv->data[j],GMP_RNDN);
	mpfr_add(tmp,tmp2,tmp,GMP_RNDN);
	}

	//last step 1/2
	mpfr_mul(tmp2,inm->data[(inm->size1-1)*(inm->memstep)+i],inv->data[(inm->size1-1)],GMP_RNDN);
	mpfr_mul(tmp3,tmp2,hlf,GMP_RNDN);
	mpfr_add(tmp,tmp3,tmp,GMP_RNDN);


	mpfr_set(out->data[i],tmp,GMP_RNDN);
 
   }
 }

	mpfr_clear(tmp);
	mpfr_clear(tmp2);
	mpfr_clear(tmp3);
	mpfr_clear(hlf);

}

void gmp_mult3m(gslgmp_matrix* in1, int t1,gslgmp_matrix* in2, int t2,gslgmp_matrix* in3,int t3, gslgmp_matrix* out)
{

 	gslgmp_matrix *tmp;
	if(t2==1&&t3==1) tmp=gslgmp_matrix_calloc(in2->size1,in3->size1);
	if(t2==1&&t3==0) tmp=gslgmp_matrix_calloc(in2->size2,in3->size2);
	if(t2==0&&t3==1) tmp=gslgmp_matrix_calloc(in2->size1,in3->size1);
	if(t2==0&&t3==0) tmp=gslgmp_matrix_calloc(in2->size1,in3->size2);

	gmp_mult2m(in2,t2,in3,t3,tmp);
	gmp_mult2m(in1,t1,tmp,0,out);

	gslgmp_matrix_free(tmp);


}

void gmp_multmDIAGm(gslgmp_matrix* in1, int t1,gslgmp_vector* in2, gslgmp_matrix* in3,int t3, gslgmp_matrix* out)
{

 	gslgmp_matrix *tmp;

	if(t3==1){ tmp=gslgmp_matrix_calloc(in2->size,in3->size1);

		  if(in3->size2!=in2->size) {fprintf(stderr,"mDiagm error 2nd Matrix and vector do not match !\n");exit(-1);}
		 }
	if(t3==0){ tmp=gslgmp_matrix_calloc(in2->size,in3->size2);

		  if(in3->size1!=in2->size) {fprintf(stderr,"mDiagm error 2nd Matrix and vector do not match !\n");exit(-1);}
		}

	mpfr_t tmp1;
	mpfr_init2(tmp1,PRECBITS);
	mpfr_t tmp2;
	mpfr_init2(tmp2,PRECBITS);
	

	if(t3==0){

		for(int i=0;i<in3->size1;i++){

		   gslgmp_vector_get(in2,i,tmp1);

		   for(int j=0;j<in3->size2;j++){
	
		   mpfr_mul(tmp->data[i*(tmp->memstep)+j],tmp1,in3->data[i*(in3->memstep)+j],GMP_RNDN);

		}}

	}

	if(t3==1){

		for(int i=0;i<in3->size2;i++){

		   gslgmp_vector_get(in2,i,tmp1);

		   for(int j=0;j<in3->size1;j++){
	
		   mpfr_mul(tmp->data[i*(tmp->memstep)+j],tmp1,in3->data[j*(in3->memstep)+i],GMP_RNDN);

		   
		}}


	}


	gmp_mult2m(in1,t1,tmp,0,out);
	mpfr_clear(tmp1);
	mpfr_clear(tmp2);

	gslgmp_matrix_free(tmp);


}


void gmp_multDIAGm(gslgmp_vector* in2, gslgmp_matrix* in3,int t3, gslgmp_matrix* out)
{

	if(t3==1){ if(out->size1!=in2->size || out->size2!=in3->size1 ) {fprintf(stderr,"Diagm error size of output matrix does not match !\n");exit(-1);}
		   if(in3->size2!=in2->size) {fprintf(stderr,"Diagm error Matrix and vector do not match !\n");exit(-1);}
		 }
	if(t3==0){ if(out->size1!=in2->size || out->size2!=in3->size2 ) {fprintf(stderr,"Diagm error size of output matrix does not match !\n");exit(-1);}
		   if(in3->size1!=in2->size) {fprintf(stderr,"Diagm error Matrix and vector do not match !\n");exit(-1);}
		}

	mpfr_t tmp1;
	mpfr_init2(tmp1,PRECBITS);
	mpfr_t tmp2;
	mpfr_init2(tmp2,PRECBITS);
	

	if(t3==0){

		for(int i=0;i<in3->size1;i++){

		   gslgmp_vector_get(in2,i,tmp1);

		   for(int j=0;j<in3->size2;j++){
	
		   mpfr_mul(out->data[i*(out->memstep)+j],tmp1,in3->data[i*(in3->memstep)+j],GMP_RNDN);

		}}

	}

	if(t3==1){

		for(int i=0;i<in3->size2;i++){

		   gslgmp_vector_get(in2,i,tmp1);

		   for(int j=0;j<in3->size1;j++){
	
		   mpfr_mul(out->data[i*(out->memstep)+j],tmp1,in3->data[j*(in3->memstep)+i],GMP_RNDN);

		   
		}}


	}

	mpfr_clear(tmp1);
	mpfr_clear(tmp2);

}

void gmp_mult4m(gslgmp_matrix* in1, int t1,gslgmp_matrix* in2, int t2,gslgmp_matrix* in3,int t3,gslgmp_matrix* in4,int t4, gslgmp_matrix* out)
{

 	gslgmp_matrix *tmp;
	if(t3==1&&t4==1) tmp=gslgmp_matrix_calloc(in3->size1,in4->size1);
	if(t3==1&&t4==0) tmp=gslgmp_matrix_calloc(in3->size2,in4->size2);
	if(t3==0&&t4==1) tmp=gslgmp_matrix_calloc(in3->size1,in4->size1);
	if(t3==0&&t4==0) tmp=gslgmp_matrix_calloc(in3->size1,in4->size2);

	gmp_mult2m(in3,t3,in4,t4,tmp);
	gmp_mult3m(in1,t1,in2,t2,tmp,0,out);

	gslgmp_matrix_free(tmp);


}

void gmp_mult5m(gslgmp_matrix* in1, int t1,gslgmp_matrix* in2, int t2,gslgmp_matrix* in3,int t3,gslgmp_matrix* in4,int t4,gslgmp_matrix* in5,int t5 ,gslgmp_matrix* out)
{

 	gslgmp_matrix *tmp;

	if(t5==1&&t4==1) tmp=gslgmp_matrix_calloc(in3->size2,in5->size1);
	if(t5==0&&t4==1) tmp=gslgmp_matrix_calloc(in3->size1,in5->size1);
	if(t5==1&&t4==0) tmp=gslgmp_matrix_calloc(in3->size2,in5->size2);
	if(t5==0&&t4==0) tmp=gslgmp_matrix_calloc(in3->size1,in5->size2);

	gmp_mult3m(in3,t3,in4,t4,in5,t5,tmp);
	gmp_mult3m(in1,t1,in2,t2,tmp,0,out);

	gslgmp_matrix_free(tmp);

}

void gmp_multmvmt(gslgmp_matrix* in1,gslgmp_vector* in2,gslgmp_matrix* in3,gslgmp_matrix* out)
{

	gslgmp_matrix* tmp;
	tmp=gslgmp_matrix_calloc(in2->size,in2->size);
	gslgmp_matrix* tmp2;
	tmp2=gslgmp_matrix_calloc(in2->size,in3->size2);

	for(int i=0;i<in2->size;i++) mpfr_set(tmp->data[i*(tmp->memstep)+i],in2->data[i],GMP_RNDN);

	gmp_mult2m(tmp,0,in3,1,tmp2);
	gmp_mult2m(in1,0,tmp2,0,out);

	gslgmp_matrix_free(tmp);
	gslgmp_matrix_free(tmp2);



}


//**********************************************************************************
//  GMP AUXILIARY VECTOR AND MATRIX ROUTINES
//**********************************************************************************

int gslgmp_vector_max_index(gslgmp_vector* in)
{
	int pos=0;
   	mpfr_t tmp;
   	mpfr_init2(tmp,PRECBITS);
	mpfr_set_inf (tmp,-1);

	for(int i=0;i<in->size;i++){
		if(mpfr_cmp(in->data[i],tmp)>=0){
		mpfr_set(tmp,in->data[i],GMP_RNDN);
		pos=i;
		}
	}

	mpfr_clear(tmp);
	return pos;
}

void gslgmp_norm (gslgmp_vector* in, mpfr_t out)
{
 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_set_d(tmp,0.,GMP_RNDN);

 mpfr_t tmp2;
 mpfr_init2(tmp2,PRECBITS);
 

 for(int i=0; i<in->size;i++){
 mpfr_mul(tmp2,in->data[i],in->data[i],GMP_RNDN);
 mpfr_add(tmp,tmp2,tmp,GMP_RNDN);
 }

 mpfr_sqrt(out,tmp,GMP_RNDN);

 mpfr_clear(tmp);
 mpfr_clear(tmp2);

}

double gslgmp_norm_d (gslgmp_vector* in)
{
 double out;
 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_set_d(tmp,0.,GMP_RNDN);

 mpfr_t tmp2;
 mpfr_init2(tmp2,PRECBITS);
 

 for(int i=0; i<in->size;i++){
 mpfr_mul(tmp2,in->data[i],in->data[i],GMP_RNDN);
 mpfr_add(tmp,tmp2,tmp,GMP_RNDN);
 }
 
 out=sqrt(mpfr_get_d(tmp,GMP_RNDN));
 
 mpfr_clear(tmp);
 mpfr_clear(tmp2);

 return out;

}

void gmpScalProd(gslgmp_matrix* S, gslgmp_vector* a, gslgmp_vector* b, mpfr_t out) //Calc (a,b)_S
{
	mpfr_t sum;
	mpfr_init2(sum,PRECBITS);
	mpfr_t ptmp1;
	mpfr_init2(ptmp1,PRECBITS);
	mpfr_t ptmp2;
	mpfr_init2(ptmp2,PRECBITS);
	mpfr_t ptmp3;
	mpfr_init2(ptmp3,PRECBITS);

	mpfr_set_d(sum,0.,GMP_RNDN);

       	gslgmp_vector* tmp;
	tmp=gslgmp_vector_calloc(S->size1);

	gmp_multmv(S,0,b,tmp);

	for(int i=0;i<S->size1;i++){
	gslgmp_vector_get(a,i,ptmp1);
	gslgmp_vector_get(tmp,i,ptmp2);
	mpfr_mul(ptmp3,ptmp1,ptmp2,GMP_RNDN);

	mpfr_set(ptmp1,sum,GMP_RNDN);
	mpfr_add(sum,ptmp1,ptmp3,GMP_RNDN);

	}
	mpfr_set(out,sum,GMP_RNDN);


	gslgmp_vector_free(tmp);
	mpfr_clear(sum);
	mpfr_clear(ptmp1);
	mpfr_clear(ptmp2);
	mpfr_clear(ptmp3);


}

void gslgmp_ddot(gslgmp_vector* in1, gslgmp_vector* in2, mpfr_t out)
{

 if(in1->size!=in2->size){fprintf(stderr,"Vector size for scalar product does not match! exiting\n");exit(-1);}

 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_set_d(tmp,0.,GMP_RNDN);

 mpfr_t tmp2;
 mpfr_init2(tmp2,PRECBITS);
 mpfr_set_d(tmp2,0.,GMP_RNDN);

 for(int i=0; i<in1->size;i++){
 mpfr_mul(tmp,in1->data[i],in2->data[i],GMP_RNDN);
 mpfr_add(tmp2,tmp,tmp2,GMP_RNDN);
 }

 mpfr_set(out,tmp2,GMP_RNDN);

 mpfr_clear(tmp);
 mpfr_clear(tmp2);

}

double gslgmp_ddot_d(gslgmp_vector* in1, gslgmp_vector* in2)
{
 double out;
 if(in1->size!=in2->size){fprintf(stderr,"Vector size for scalar product does not match! exiting\n");exit(-1);}

 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_set_d(tmp,0.,GMP_RNDN);

 mpfr_t tmp2;
 mpfr_init2(tmp2,PRECBITS);
 mpfr_set_d(tmp2,0.,GMP_RNDN);

 for(int i=0; i<in1->size;i++){
 mpfr_mul(tmp,in1->data[i],in2->data[i],GMP_RNDN);
 mpfr_add(tmp2,tmp,tmp2,GMP_RNDN);
 }

 out=mpfr_get_d(tmp2,GMP_RNDN);

 mpfr_clear(tmp);
 mpfr_clear(tmp2);

 return out;

}

void gmp_hypot(mpfr_t a, mpfr_t b,mpfr_t out)
{

 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_set_d(tmp,0.,GMP_RNDN);

 mpfr_t tmp1;
 mpfr_init2(tmp1,PRECBITS);
 mpfr_set_d(tmp1,0.,GMP_RNDN);

 mpfr_t tmp2;
 mpfr_init2(tmp2,PRECBITS);
 mpfr_set_d(tmp2,0.,GMP_RNDN);

 mpfr_mul(tmp1,a,a,GMP_RNDN);
 mpfr_mul(tmp2,b,b,GMP_RNDN);

 mpfr_add(tmp,tmp1,tmp2,GMP_RNDN);

 mpfr_sqrt(out,tmp,GMP_RNDN);

 mpfr_clear(tmp);
 mpfr_clear(tmp1);
 mpfr_clear(tmp2);

}

void gmpabs(mpfr_t in, mpfr_t out)
{
 mpfr_t tmp;
 mpfr_init2(tmp,PRECBITS);
 mpfr_set_d(tmp,0.,GMP_RNDN);

 if(mpfr_cmp(in,tmp)>=0)
 {
  mpfr_set(out,in,GMP_RNDN);
 }else{
  mpfr_neg(out,in,GMP_RNDN);
 }

 mpfr_clear(tmp);

}


//**********************************************************************************
//  GMP VECTOR AND MATRIX ROUTINES TRANSFERRING MEMORY FROM, e.g.
//		MATRIX -> SUBMATRIX
//		VECTOR -> SUBVECTOR
//	   MATRIX(COL) -> VECTOR
//		VECTOR -> MATRIX
//**********************************************************************************

 void gslgmp_matrix_memcpy(gslgmp_matrix* out , gslgmp_matrix* in)
{

	for(int i=0;i<in->size1*in->size2;i++){

	mpfr_set(out->data[i],in->data[i],GMP_RNDN);

	}

}

void gslgmp_submatrix (gslgmp_matrix* A, int s1,int s2, int size1, int size2, gslgmp_matrix* out)
{

	int ci=0;
	int cj=0;

	for(int i=s1;i<s1+size1;i++){
	   for(int j=s2;j<s2+size2;j++){

	mpfr_set(out->data[ci*(size2)+cj],A->data[i*(A->size2)+j],GMP_RNDN); cj++;

	}ci++;cj=0;}
}

void gslgmp_reverse_submatrix (gslgmp_matrix* A, int s1,int s2, int size1, int size2, gslgmp_matrix* out)
{

	int ci=0;
	int cj=0;

	for(int i=s1;i<s1+size1;i++){
	   for(int j=s2;j<s2+size2;j++){

	mpfr_set(A->data[i*(A->size2)+j],out->data[ci*(size2)+cj],GMP_RNDN); cj++;

	}ci++;cj=0;}
}

void gslgmp_matrix_column (gslgmp_matrix* A, int j, gslgmp_vector* out)
{
	if(A->size1!=out->size){fprintf(stderr,"column dimension od not match, exiting\n");exit(-1);}

	for(int i=0;i<A->size1;i++){
	mpfr_set(out->data[i],A->data[i*(A->size2)+j],GMP_RNDN);
	}
}

void gslgmp_matrix_reverse_column (gslgmp_matrix* A, int j, gslgmp_vector* out)
{
	if(A->size1!=out->size){fprintf(stderr,"column dimension od not match, exiting\n");exit(-1);}

	for(int i=0;i<A->size1;i++){
	mpfr_set(A->data[i*(A->size2)+j],out->data[i],GMP_RNDN);
	}
}

void gslgmp_matrix_subcolumn (gslgmp_matrix* A, int j, int offset, int len, gslgmp_vector* out)
{
	if(out->size!=len){fprintf(stderr,"column dimension do not match, exiting\n");exit(-1);}

	for(int i=0;i<len;i++){
	mpfr_set(out->data[i],A->data[(i+offset)*(A->size2)+j],GMP_RNDN);
	}
}

void gslgmp_matrix_reverse_subcolumn (gslgmp_matrix* A, int j, int offset, int len, gslgmp_vector* out)
{
	if(out->size!=len){fprintf(stderr,"column dimension do not match, exiting\n");exit(-1);}

	for(int i=0;i<len;i++){
	mpfr_set(A->data[(i+offset)*(A->size2)+j],out->data[i],GMP_RNDN);
	}
}


void gslgmp_vector_part (gslgmp_vector* in, int offset, int size, gslgmp_vector* out)
{
	if(out->size!=(size)){fprintf(stderr,"Vectors do not match\n");exit(-1);}

	for(int i=offset;i<offset+size;i++){
	mpfr_set(out->data[i-offset],in->data[i],GMP_RNDN);
	}
}

void gslgmp_vector_partN (gslgmp_vector* in, int N, int offset, int size, gslgmp_vector* out)
{
	int limit;

	if(out->size!=(size)){fprintf(stderr,"Vectors do not match\n");exit(-1);}

	if(offset+((size>N)?N:size)>in->size)
	{ limit=in->size;
	}else{
	  limit=offset+((size>N)?N:size);
	}

	for(int i=offset;i<limit;i++){
	mpfr_set(out->data[i-offset],in->data[i],GMP_RNDN);
	}
}


void gslgmp_vector_reverse_part (gslgmp_vector* in, int offset, int size, gslgmp_vector* out)
{
	if(out->size!=(size)){fprintf(stderr,"Vectors do not match\n");exit(-1);}

	for(int i=offset;i<offset+size;i++){
	mpfr_set(in->data[i],out->data[i-offset],GMP_RNDN);
	}
}

void gslgmp_vector_reverse_partN (gslgmp_vector* in,int N, int offset, int size, gslgmp_vector* out)
{
	int limit;
	if(out->size!=(size)){fprintf(stderr,"Vectors do not match\n");exit(-1);}

	if(offset+((size>N)?N:size)>in->size)
	{ limit=in->size;
	}else{
	  limit=offset+((size>N)?N:size);
	}

	for(int i=offset;i<limit;i++){
	mpfr_set(in->data[i],out->data[i-offset],GMP_RNDN);
	}
}
void gmp_VMakeM(gslgmp_vector* in,gslgmp_matrix* out)
{
	if(out->size1!=in->size||out->size2!=in->size){fprintf(stderr,"vector and matrix incompatible\n");exit(-1);}

	for(int i=0;i<in->size*in->size;i++) mpfr_set_d(out->data[i],0.0,GMP_RNDN);

	for(int i=0;i<in->size;i++) mpfr_set(out->data[i*(out->size2)+i],in->data[i],GMP_RNDN);
}

 void gslgmp_vector_memcpy(gslgmp_vector* out , gslgmp_vector* in)
{

	for(int i=0;i<in->size;i++){

	mpfr_set(out->data[i],in->data[i],GMP_RNDN);

	}

}


//**********************************************************************************
//  GMP LINEAR ALGEBRA ROUTINES THAT OVERALL PROVIDE THE NECESSARY
//		- SVD Decomposition (also of the transpose)
//		- QR decomposition
//		- QR solver
//		- QR Inverter
//	using:
//		- Housholder transforms
//		- BLAS solver routine
//**********************************************************************************

#define GMP_MAX(XX,YY) ((XX>YY)?XX:YY)

void gmpSVD_internal (gslgmp_matrix * A, gslgmp_matrix * Q, gslgmp_vector * S, int* NSING)
{
  mpfr_t myprec;
  mpfr_init2(myprec,PRECBITS);
  mpfr_ui_pow_ui(myprec,2,PRECBITS,GMP_RNDN);
  mpfr_ui_div(myprec,1,myprec,GMP_RNDN);

    
  if (A->size1 < A->size2)
    {
      /* FIXME: only implemented  M>=N case so far */

      fprintf (stderr,"svd of MxN matrix, M<N, is not implemented");
    }
  else if (Q->size1 != A->size2)
    {
      fprintf (stderr,"square matrix Q must match second dimension of matrix A");
    }
  else if (Q->size1 != Q->size2)
    {
      fprintf (stderr,"matrix Q must be square");
    }
  else if (S->size != A->size2)
    {
      fprintf (stderr,"length of vector S must match second dimension of matrix A");
    }
  else
    {
      const int M = A->size1;
      const int N = A->size2;
      int i, j, k;

      /* Initialize the rotation counter and the sweep counter. */
      int count = 1;
      int sweep = 0;
      int sweepmax = 5*N;

      mpfr_t tolerance;
      mpfr_init2(tolerance,PRECBITS) ;
      mpfr_set_d(tolerance,10. * (double)M ,GMP_RNDN); 
      mpfr_mul(tolerance,tolerance,myprec,GMP_RNDN);
      /* Always do at least 12 sweeps. */
      sweepmax = GMP_MAX (sweepmax, 12);

      /* Set Q to the identity matrix. */
      gslgmp_matrix_set_identity (Q);




      /* Store the column error estimates in S, for use during the
         orthogonalization */


      gslgmp_vector* column;
      column=gslgmp_vector_calloc(A->size1);
      mpfr_t tm1;
      mpfr_init2(tm1,PRECBITS);

      mpfr_t sj;
      mpfr_init2(sj,PRECBITS);


      for (j = 0; j < N; j++)
        {
 	  gslgmp_matrix_column (A,j,column);

  	  gslgmp_norm(column,sj);

	  mpfr_set(tm1,myprec,GMP_RNDN);
	  mpfr_mul(sj,sj,tm1,GMP_RNDN);	

	  gslgmp_vector_set(S,j,sj);
        }

      /* Orthogonalize A by plane rotations. */

        int sorted, orthog, noisya, noisyb;
	mpfr_t a,b,p,q,cosine,sine,v,abserr_a,abserr_b,two,tmp1,tmp2,tmp3,Aik,Aij,Qij,Qik,prev_norm,norm,zero;

	mpfr_init2(prev_norm,PRECBITS);
	mpfr_init2(norm,PRECBITS);
	mpfr_init2(Qik,PRECBITS);
	mpfr_init2(Qij,PRECBITS);
	mpfr_init2(Aik,PRECBITS);
	mpfr_init2(Aij,PRECBITS);
	mpfr_init2(a,PRECBITS);
	mpfr_init2(b,PRECBITS);
	mpfr_init2(p,PRECBITS);
	mpfr_init2(q,PRECBITS);
	mpfr_init2(v,PRECBITS);
	mpfr_init2(two,PRECBITS);
	mpfr_init2(zero,PRECBITS);
	mpfr_init2(tmp1,PRECBITS);
	mpfr_init2(tmp2,PRECBITS);
	mpfr_init2(tmp3,PRECBITS);
	mpfr_init2(sine,PRECBITS);
	mpfr_init2(cosine,PRECBITS);
	mpfr_init2(abserr_a,PRECBITS);
	mpfr_init2(abserr_b,PRECBITS);

	mpfr_set_d(prev_norm,0.0,GMP_RNDN);
	mpfr_set_d(norm,0.0,GMP_RNDN);
	mpfr_set_d(Qik,0.0,GMP_RNDN);
	mpfr_set_d(Qij,0.0,GMP_RNDN);
	mpfr_set_d(Aik,0.0,GMP_RNDN);
	mpfr_set_d(Aij,0.0,GMP_RNDN);
	mpfr_set_d(a,0.0,GMP_RNDN);
	mpfr_set_d(v,0.0,GMP_RNDN);
	mpfr_set_d(b,0.0,GMP_RNDN);
	mpfr_set_d(p,0.0,GMP_RNDN);
	mpfr_set_d(q,0.0,GMP_RNDN);
	mpfr_set_d(two,2.0,GMP_RNDN);
	mpfr_set_d(zero,0.0,GMP_RNDN);
	mpfr_set_d(tmp1,0.0,GMP_RNDN);
	mpfr_set_d(tmp2,0.0,GMP_RNDN);

	gslgmp_vector *cj,*ck;

	cj=gslgmp_vector_calloc(A->size1);
	ck=gslgmp_vector_calloc(A->size1);

	
      while (count > 0 && sweep <= sweepmax)
        {
          /* Initialize rotation counter. */
          count = N * (N - 1) / 2;



          for (j = 0; j < N - 1; j++)
            {

              for (k = j + 1; k < N; k++)
                {
			
		   gslgmp_matrix_column (A,j,cj);
		   gslgmp_matrix_column (A,k,ck);

		  gslgmp_ddot (cj, ck,p);

		  mpfr_mul(p,p,two,GMP_RNDN);

		  gslgmp_norm(cj,a);
		  gslgmp_norm(ck,b);


		  mpfr_mul(tmp1,a,a,GMP_RNDN);
		  mpfr_mul(tmp2,b,b,GMP_RNDN);
		  mpfr_sub(q,tmp1,tmp2,GMP_RNDN);
		
		  gmp_hypot(p,q,v);

		  /* test for columns j,k orthogonal, or dominant errors */

		  gslgmp_vector_get(S,j,abserr_a);
		  gslgmp_vector_get(S,k,abserr_b);


		  sorted=((mpfr_cmp(a,b)>=0)?1:0);

		 mpfr_mul(tmp1,a,b,GMP_RNDN);
		 mpfr_mul(tmp2,tmp1,tolerance,GMP_RNDN);
		 gmpabs(p,tm1);
	         orthog=((mpfr_cmp(tm1,tmp2)<=0)?1:0);
				 
		  noisya=((mpfr_cmp(a,abserr_a)<0)?1:0);
		  noisyb=((mpfr_cmp(b,abserr_b)<0)?1:0);


                  if (sorted && (orthog || noisya || noisyb))
                    {
                      count--;
                      continue;
                    }

                  /* calculate rotation angles */
                  if (v == 0 || !sorted)
                    {
                      mpfr_set_d(cosine,0.0,GMP_RNDN);
                      mpfr_set_d(sine,1.0,GMP_RNDN);
                    }
                  else
                    {
		     mpfr_mul(tmp1,two,v,GMP_RNDN);
		     mpfr_add(tmp2,v,q,GMP_RNDN);
		     mpfr_div(tmp3,tmp2,tmp1,GMP_RNDN);
		     mpfr_sqrt(cosine,tmp3,GMP_RNDN);

		     mpfr_mul(tmp1,two,v,GMP_RNDN);
		     mpfr_mul(tmp2,tmp1,cosine,GMP_RNDN);
		     mpfr_div(sine,p,tmp2,GMP_RNDN);

                    }

                  /* apply rotation to A */
                  for (i = 0; i < M; i++)
                    {
			gslgmp_matrix_get (A, i, k, Aik);
			gslgmp_matrix_get (A, i, j, Aij);

		      mpfr_mul(tmp1,Aik,sine,GMP_RNDN);
		      mpfr_mul(tmp2,Aij,cosine,GMP_RNDN);
		      mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
		      gslgmp_matrix_set(A,i,j,tmp3);

		      mpfr_mul(tmp1,Aik,cosine,GMP_RNDN);
		      mpfr_mul(tmp3,Aij,sine,GMP_RNDN);
		      mpfr_neg(tmp2,tmp3,GMP_RNDN);


		      mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
		      gslgmp_matrix_set(A,i,k,tmp3);
		      

                    }
		    
			gmpabs(cosine,tm1);
		    mpfr_mul(tmp1,tm1,abserr_a,GMP_RNDN);
			gmpabs(sine,tm1);
		    mpfr_mul(tmp2,tm1,abserr_b,GMP_RNDN);
		    gslgmp_vector_set(S,j,tmp2);


			gmpabs(sine,tm1);
		    mpfr_mul(tmp1,tm1,abserr_a,GMP_RNDN);
			gmpabs(cosine,tm1);
		    mpfr_mul(tmp2,tm1,abserr_b,GMP_RNDN);
		    gslgmp_vector_set(S,k,tmp2);


                  /* apply rotation to Q */
                  for (i = 0; i < N; i++)
                    {

			gslgmp_matrix_get (Q, i, j,Qij);
			gslgmp_matrix_get (Q, i, k,Qik);

		      mpfr_mul(tmp1,Qik,sine,GMP_RNDN);
		      mpfr_mul(tmp2,Qij,cosine,GMP_RNDN);
		      mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
		      gslgmp_matrix_set(Q,i,j,tmp3);


		      mpfr_mul(tmp1,Qik,cosine,GMP_RNDN);
		      mpfr_mul(tmp2,Qij,sine,GMP_RNDN);
		      mpfr_sub(tmp3,tmp1,tmp2,GMP_RNDN);
		      gslgmp_matrix_set(Q,i,k,tmp3);

                    }
                }
            }

          /* Sweep completed. */
          sweep++;
        }

      /* 
       * Orthogonalization complete. Compute singular values.
       */

      {
        mpfr_set_d(prev_norm,-1.0,GMP_RNDN);

        for (j = 0; j < N; j++)
          {

	    gslgmp_matrix_column (A,j,column);
	    gslgmp_norm(column,norm);

            /* Determine if singular value is zero, according to the
               criteria used in the main loop above (i.e. comparison
               with norm of previous column). */

		mpfr_mul(tmp1,tolerance,prev_norm,GMP_RNDN);

            if (mpfr_cmp(norm,zero) == 0 || mpfr_cmp(prev_norm,zero) == 0 
                || (j > 0 && mpfr_cmp(norm,tmp1) <= 0))
              {
                gslgmp_vector_set_d (S, j, 0.0);     /* singular */

		for(int l=0;l<A->size1;l++){
		gslgmp_matrix_set_d(A,l,j,0.0);
		}


                mpfr_set_d(prev_norm,0.0,GMP_RNDN);
              }
            else
              {
                gslgmp_vector_set (S, j, norm);    /* non-singular */

		for(int l=0;l<A->size1;l++){
		gslgmp_matrix_get(A,l,j,tmp1);
		mpfr_ui_div(tmp2,1,norm,GMP_RNDN);
		mpfr_mul(tmp3,tmp1,tmp2,GMP_RNDN);
		gslgmp_matrix_set(A,l,j,tmp3);

		}


                mpfr_set(prev_norm,norm,GMP_RNDN);
              }
          }
      }

	for(int i=0;i<S->size;i++){

	if(gslgmp_vector_get_d(S,i)>CONDNSING) *NSING=(i+1);

	}


      if (count > 0)
        {
          /* reached sweep limit */
          fprintf(stderr,"Jacobi iterations did not reach desired tolerance\n");
        }

	gslgmp_vector_free(column);
	mpfr_clear(a);
	mpfr_clear(b);
	mpfr_clear(p);
	mpfr_clear(q);
	mpfr_clear(cosine);
	mpfr_clear(sine);
	mpfr_clear(v);
	mpfr_clear(abserr_a);
	mpfr_clear(abserr_b);
	mpfr_clear(two);
	mpfr_clear(tmp1);
	mpfr_clear(tmp2);
	mpfr_clear(tmp3);
	mpfr_clear(Aik);
	mpfr_clear(Aij);
	mpfr_clear(Qij);
	mpfr_clear(Qik);
	mpfr_clear(prev_norm);
	mpfr_clear(norm);
	mpfr_clear(zero);

    }


  mpfr_clear(myprec);

}

void gmpSVD(gslgmp_matrix * in, gslgmp_matrix * U, gslgmp_matrix * V, gslgmp_vector * S, int* NSING)
{

	gslgmp_matrix* tmp;
	tmp=gslgmp_matrix_calloc(in->size1,in->size2);

	for(int i=0;i<in->size1*in->size2;i++){
	mpfr_set(tmp->data[i],in->data[i],GMP_RNDN);
	}

	gmpSVD_internal (tmp, V, S, NSING);

	for(int i=0;i<in->size1*in->size2;i++){
	mpfr_set(U->data[i],tmp->data[i],GMP_RNDN);
	}
	
	gslgmp_matrix_free(tmp);

}

void gmpSVD_Transpose(gslgmp_matrix * in, gslgmp_matrix * U, gslgmp_matrix * V, gslgmp_vector * S, int* NSING)
{

	gslgmp_matrix* tmp;
	tmp=gslgmp_matrix_calloc(in->size2,in->size1);

	for(int i=0;i<in->size1;i++){
	   for(int j=0;j<in->size2;j++){
	mpfr_set(tmp->data[j*tmp->size2+i],in->data[i*in->size2+j],GMP_RNDN);
	}}

	gmpSVD_internal (tmp, V, S, NSING);

	for(int i=0;i<in->size1*in->size2;i++){
	mpfr_set(U->data[i],tmp->data[i],GMP_RNDN);
	}
	
	gslgmp_matrix_free(tmp);

}

void
gslgmp_linalg_householder_transform (gslgmp_vector * v,mpfr_t outtau)
{
  /* replace v[0:n-1] with a householder vector (v[0:n-1]) and
     coefficient tau that annihilate v[1:n-1] */

  const int n = v->size ;
  gslgmp_vector* tmp;
      mpfr_t alpha,beta,tau,xnorm,zero,tmpv,tmpv2,myprec;
      mpfr_init2(alpha,PRECBITS);
      mpfr_init2(beta,PRECBITS);
      mpfr_init2(tau,PRECBITS);
      mpfr_init2(xnorm,PRECBITS);
      mpfr_init2(zero,PRECBITS);
      mpfr_init2(tmpv,PRECBITS);
      mpfr_init2(tmpv2,PRECBITS);
      mpfr_set_ui(zero,0,GMP_RNDN);
      mpfr_init2(myprec,PRECBITS);
      mpfr_ui_pow_ui(myprec,2,PRECBITS,GMP_RNDN);
      mpfr_ui_div(myprec,1,myprec,GMP_RNDN);

      

  if (n == 1)
    {
      mpfr_set_ui(outtau, 0,GMP_RNDN); /* tau = 0 */

      mpfr_clear(alpha);
      mpfr_clear(tau);
      mpfr_clear(xnorm);
      mpfr_clear(zero);
      mpfr_clear(tmpv);
      mpfr_clear(tmpv2);
      mpfr_clear(beta);
      mpfr_clear(myprec);
	

      return;
    }
  else
    { 

     
      tmp=gslgmp_vector_calloc(n-1);

      gslgmp_vector_part(v,1,n-1,tmp);

      gslgmp_norm(tmp,xnorm);

      
      
      if ( mpfr_less_p(xnorm,myprec) ) 
        {
           mpfr_set_ui(outtau, 0,GMP_RNDN);

	   gslgmp_vector_free(tmp);

      mpfr_clear(alpha);
      mpfr_clear(tau);
      mpfr_clear(xnorm);
      mpfr_clear(myprec);
      mpfr_clear(zero);
      mpfr_clear(tmpv);
      mpfr_clear(tmpv2);
      mpfr_clear(beta);
	   return;
        }
      
        gslgmp_vector_get(v,0,alpha);
	gmp_hypot(alpha,xnorm,tmpv);

      if(mpfr_greaterequal_p(alpha,myprec))
	{ mpfr_neg(beta,tmpv,GMP_RNDN);
	}else{
	  mpfr_set(beta,tmpv,GMP_RNDN);
	}

      mpfr_sub(tmpv,beta,alpha,GMP_RNDN);
      mpfr_div(tau,tmpv,beta,GMP_RNDN);

      for(int i=0;i<tmp->size;i++){

	mpfr_sub(tmpv,alpha,beta,GMP_RNDN);
	mpfr_ui_div(tmpv2,1,tmpv,GMP_RNDN);
	mpfr_mul(tmpv,v->data[1+i],tmpv2,GMP_RNDN); //START VALUE IS ALWAYS 1

	mpfr_set(v->data[1+i],tmpv,GMP_RNDN);
	
	}



      gslgmp_vector_set (v, 0, beta) ;
      mpfr_set(outtau,tau,GMP_RNDN);

      gslgmp_vector_free(tmp);
   }

      mpfr_clear(alpha);
      mpfr_clear(tau);
      mpfr_clear(xnorm);
      mpfr_clear(myprec);
      mpfr_clear(zero);
      mpfr_clear(tmpv);
      mpfr_clear(tmpv2);
      mpfr_clear(beta);

}


void
gslgmp_linalg_householder_hm (mpfr_t tau, gslgmp_vector * v, gslgmp_matrix * A)
{

/*	fprintf(stderr,"gmpTAU: %g\n",mpfr_get_d(tau,GMP_RNDN));
 */ /* applies a householder transformation v,tau to matrix m */
	mpfr_t zero,tmp1,tmp2,tmp3,wj,A0j,Aij,vi,myprec;

	mpfr_init2(zero,PRECBITS);
	mpfr_init2(wj,PRECBITS);
	mpfr_init2(tmp1,PRECBITS);
	mpfr_init2(tmp2,PRECBITS);
	mpfr_init2(tmp3,PRECBITS);
	mpfr_init2(A0j,PRECBITS);
	mpfr_init2(Aij,PRECBITS);
	mpfr_init2(vi,PRECBITS);

	mpfr_set_ui(zero,0,GMP_RNDN);

	mpfr_init2(myprec,PRECBITS);
	mpfr_ui_pow_ui(myprec,2,PRECBITS,GMP_RNDN);
	mpfr_ui_div(myprec,1,myprec,GMP_RNDN);
	
	mpfr_abs(tmp1,tau,GMP_RNDN);
	
   if ( mpfr_less_p(tmp1,myprec) )
    {
   	mpfr_clear(zero);
   	mpfr_clear(tmp1);
   	mpfr_clear(tmp2);
   	mpfr_clear(tmp3);
   	mpfr_clear(myprec);
   	mpfr_clear(wj);
   	mpfr_clear(A0j);
   	mpfr_clear(Aij);
   	mpfr_clear(vi);

      return;
    }

  
    int i, j;
    
    for (j = 0; j < A->size2; j++)
      {
        /* Compute wj = Akj vk */
        
        gslgmp_matrix_get(A,0,j,wj);  
       
        for (i = 1; i < A->size1; i++)  /* note, computed for v(0) = 1 above */
          {
		gslgmp_matrix_get(A,i,j,tmp1);
		gslgmp_vector_get(v,i,tmp2);

		mpfr_mul(tmp3,tmp1,tmp2,GMP_RNDN);
		mpfr_add(tmp1,tmp3,wj,GMP_RNDN);
		mpfr_set(wj,tmp1,GMP_RNDN);
          }

 
           gslgmp_matrix_get (A, 0, j,A0j);

	  mpfr_mul(tmp1,tau,wj,GMP_RNDN);

	  mpfr_sub(tmp2,A0j,tmp1,GMP_RNDN);
         gslgmp_matrix_set (A, 0, j, tmp2);

        
       	


        /* i = 1 .. M-1 */
        
        for (i = 1; i < A->size1; i++)
          {
            gslgmp_matrix_get (A, i, j,Aij);
            gslgmp_vector_get (v, i,vi);

	  mpfr_mul(tmp1,tau,wj,GMP_RNDN);
	  mpfr_mul(tmp2,tmp1,vi,GMP_RNDN);
	  mpfr_sub(tmp3,Aij,tmp2,GMP_RNDN);

            gslgmp_matrix_set (A, i, j, tmp3);
          }
      }
  

   mpfr_clear(zero);
   mpfr_clear(tmp1);
   mpfr_clear(tmp2);
   mpfr_clear(tmp3);
   mpfr_clear(myprec);
   mpfr_clear(wj);
   mpfr_clear(A0j);
   mpfr_clear(Aij);
   mpfr_clear(vi);

  

}

void gslgmp_linalg_householder_mh (mpfr_t tau, gslgmp_vector * v, gslgmp_matrix * A)
{
  /* applies a householder transformation v,tau to matrix m from the
     right hand side in order to zero out rows */
	mpfr_t myprec;
	mpfr_init2(myprec,PRECBITS);
	mpfr_ui_pow_ui(myprec,2,PRECBITS,GMP_RNDN);
	mpfr_ui_div(myprec,1,myprec,GMP_RNDN);
 
	
	mpfr_t wi,Ai0,Aij,vj,tmp1,tmp2;

	mpfr_init2(wi,PRECBITS);
	mpfr_init2(tmp1,PRECBITS);
	mpfr_init2(tmp2,PRECBITS);
	mpfr_init2(Ai0,PRECBITS);
	mpfr_init2(Aij,PRECBITS);
	mpfr_init2(vj,PRECBITS);

	
	mpfr_abs(tmp1,tau,GMP_RNDN);
//	if (mpfr_sgn(tau) == 0){
	if ( mpfr_less_p(tmp1,myprec) ){
	    
	    mpfr_clear(tmp1);
	    mpfr_clear(tmp2);
	    mpfr_clear(myprec);
	    mpfr_clear(Aij);
	    mpfr_clear(Ai0);
	    mpfr_clear(vj);
	    mpfr_clear(wi);
	    return;

	    
	}
  
  /* A = A - tau w v' */

    int i, j;
    
    for (i = 0; i < A->size1; i++)
      {
        gslgmp_matrix_get(A,i,0,wi);  
        
        for (j = 1; j < A->size2; j++)  /* note, computed for v(0) = 1 above */
          {
		gslgmp_matrix_get(A,i,j,tmp1);
		mpfr_mul(tmp2,tmp1,v->data[j],GMP_RNDN);
		mpfr_add(wi,wi,tmp2,GMP_RNDN);
		
            //wi += gsl_matrix_get(A,i,j) * gsl_vector_get(v,j);
          }
        
        /* j = 0 */
        
        {
          gslgmp_matrix_get (A, i, 0,Ai0);
	  mpfr_mul(tmp1,tau,wi,GMP_RNDN);
	  mpfr_sub(tmp2,Ai0,tmp1,GMP_RNDN);
          gslgmp_matrix_set (A, i, 0, tmp2);
          //gsl_matrix_set (A, i, 0, Ai0 - tau *  wi);
        }
        
        /* j = 1 .. N-1 */
        
        for (j = 1; j < A->size2; j++) 
          {
            gslgmp_vector_get (v, j,vj);
            gslgmp_matrix_get (A, i, j,Aij);
	    mpfr_mul(tmp1,wi,vj,GMP_RNDN);
	    mpfr_mul(tmp1,tmp1,tau,GMP_RNDN);
	    mpfr_sub(tmp2,Aij,tmp1,GMP_RNDN);
	    gslgmp_matrix_set (A, i, j,tmp2);
            //gsl_matrix_set (A, i, j, Aij - tau * wi * vj);
          }
      }

 mpfr_clear(tmp1);
 mpfr_clear(tmp2);
 mpfr_clear(myprec);
 mpfr_clear(Aij);
 mpfr_clear(Ai0);
 mpfr_clear(vj);
 mpfr_clear(wi);
 


}

int
gslgmp_linalg_hessenberg_decomp(gslgmp_matrix *A, gslgmp_vector *tau)
{
  const int N = A->size1;

  if (N != A->size2)
    {
      fprintf(stderr,"Hessenberg reduction requires square matrix"); exit(-1);
    }
  else if (N != tau->size)
    {
      fprintf(stderr,"tau vector must match matrix size"); exit(-1);
    }
  else if (N < 3)
    {
      /* nothing to do */
      return 1;
    }
  else
    {
      int i;           /* looping */
      
      gslgmp_vector *c, *hv;
      
      //gsl_vector_view c,  /* matrix column */
      //                hv; /* householder vector */
      
      gslgmp_matrix *m;
      
      //gsl_matrix_view m;
      mpfr_t tau_i;       /* beta in algorithm 7.4.2 */
      mpfr_init2(tau_i,PRECBITS);
      
      for (i = 0; i < N - 2; ++i)
        {
          /*
           * make a copy of A(i + 1:n, i) and store it in the section
           * of 'tau' that we haven't stored coefficients in yet
           */
  
   	   c=gslgmp_vector_calloc(N - i - 1);
           gslgmp_matrix_subcolumn(A, i, i + 1, N - i - 1,c);
           //c = gsl_matrix_subcolumn(A, i, i + 1, N - i - 1);

   	   hv=gslgmp_vector_calloc(N - (i + 1));
	   gslgmp_vector_part(tau, i + 1, N - (i + 1),hv);
	   //hv = gsl_vector_subvector(tau, i + 1, N - (i + 1));
	  
	   //gsl_vector_memcpy(&hv.vector, &c.vector);
	   for(int k=0;k<N-i-1;k++) mpfr_set(hv->data[k],c->data[k],GMP_RNDN);
	   gslgmp_vector_reverse_part(tau, i + 1, N - (i + 1),hv);
	   
          /* compute householder transformation of A(i+1:n,i) */
          gslgmp_linalg_householder_transform(hv,tau_i);
	  gslgmp_vector_reverse_part(tau, i + 1, N - (i + 1),hv);
	  
          //tau_i = gsl_linalg_householder_transform(&hv.vector);

          /* apply left householder matrix (I - tau_i v v') to A */
	  
	  m=gslgmp_matrix_calloc(N - (i + 1), N - i);
	  gslgmp_submatrix(A, i + 1, i, N - (i + 1), N - i,m);
          //m = gsl_matrix_submatrix(A, i + 1, i, N - (i + 1), N - i);

	  gslgmp_linalg_householder_hm(tau_i, hv, m);
	  gslgmp_reverse_submatrix(A, i + 1, i, N - (i + 1), N - i,m);
	  gslgmp_matrix_free(m);
	  
	  
          /* apply right householder matrix (I - tau_i v v') to A */
	  m=gslgmp_matrix_calloc(N, N - (i + 1));
	  gslgmp_submatrix(A, 0, i + 1, N, N - (i + 1),m);
 	  
          //m = gsl_matrix_submatrix(A, 0, i + 1, N, N - (i + 1));
          gslgmp_linalg_householder_mh(tau_i, hv, m);
	  gslgmp_reverse_submatrix(A, 0, i + 1, N, N - (i + 1),m);

          /* save Householder coefficient */
          gslgmp_vector_set(tau, i, tau_i);

          /*
           * store Householder vector below the subdiagonal in column
           * i of the matrix. hv(1) does not need to be stored since
           * it is always 1.
           */
	  
	  gslgmp_vector_free(c);
	  gslgmp_vector_free(hv);

	  c=gslgmp_vector_calloc(N - i - 1 -1);
          gslgmp_matrix_subcolumn(A, i, i + 1+1, N - i - 1-1,c);
  
	  
          //c = gsl_vector_subvector(&c.vector, 1, c.vector.size - 1);

    	   hv=gslgmp_vector_calloc(N - (i + 1) -1 );
	   gslgmp_vector_part(tau, i + 1+1, N - (i + 1)-1,hv);

	   for(int k=0;k<N - i - 1 -1;k++) mpfr_set(c->data[k],hv->data[k],GMP_RNDN);
	   
	   //gslgmp_vector_reverse_part(tau, i + 1+1, N - (i + 1)-1,hv);
           gslgmp_matrix_reverse_subcolumn(A, i, i + 1+1, N - i - 1-1,c);
	   
	  gslgmp_vector_free(c);
	  gslgmp_vector_free(hv);
	   
           //hv = gsl_vector_subvector(&hv.vector, 1, hv.vector.size - 1);
           //gsl_vector_memcpy(&c.vector, &hv.vector);
        }

	mpfr_clear(tau_i);

      return 1;
    }
} /* gsl_linalg_hessenberg_decomp() */

void gslgmp_linalg_QR_decomp (gslgmp_matrix * A, gslgmp_vector * tau)
{
  const int M = A->size1;
  const int N = A->size2;

  gslgmp_matrix *subm;
  gslgmp_vector *c_full,*c;
  mpfr_t tau_i;

  mpfr_init2(tau_i,PRECBITS);

  if (tau->size != GSL_MIN (M, N))
    {
      fprintf(stderr, "size of tau must be MIN(M,N)");
    }
  else
    {
      int i;

      for (i = 0; i < GSL_MIN (M, N); i++)
        {
          /* Compute the Householder transformation to reduce the j-th
             column of the matrix to a multiple of the j-th unit vector */

	  c_full=gslgmp_vector_calloc(A->size1);
	  c=gslgmp_vector_calloc(M-i);

//          gsl_vector_view c_full = gsl_matrix_column (A, i);
//          gsl_vector_view c = gsl_vector_subvector (&(c_full.vector), i, M-i);

	  gslgmp_matrix_column(A,i,c_full);
	  gslgmp_vector_part(c_full,i,M-i,c);

//          double tau_i = gsl_linalg_householder_transform (&(c.vector));

          gslgmp_linalg_householder_transform (c,tau_i);  // !!!!!


	  //Update the old A column with the new data in c

// 	gmpPVec(c);
// 	fprintf(stderr,"gmpTAU_i: %g\n",mpfr_get_d(tau_i,GMP_RNDN));


	  for(int l=0;l<c->size;l++){
		gslgmp_matrix_set(A,i+l,i,c->data[l]);
	  }


//          gsl_vector_set (tau, i, tau_i);
            gslgmp_vector_set (tau, i, tau_i);

          // Apply the transformation to the remaining columns and
          //   update the norms 

          if (i + 1 < N)
            {



//              gsl_matrix_view m = gsl_matrix_submatrix (A, i, i + 1, M - i, N - (i + 1));
	      subm=gslgmp_matrix_calloc(M-i,N - (i + 1));
              gslgmp_submatrix (A, i, i + 1, M - i, N - (i + 1),subm);

// 		gmpPMat(subm);



              gslgmp_linalg_householder_hm (tau_i, c, subm); // !!!!

	//Update the new information from the previous operation:

	      gslgmp_reverse_submatrix (A, i, i + 1, M - i, N - (i + 1),subm);


 
	      for(int l=0;l<c->size;l++){
		gslgmp_matrix_set(A,i+l,i,c->data[l]);
	      }

	
//	     gmpPMat(A);
//	     gmpPVec(c);
		
	      gslgmp_matrix_free(subm);

            }
	 
	 gslgmp_vector_free(c_full);
	 gslgmp_vector_free(c);
        }


    }

  mpfr_clear(tau_i);

}


void
gmpcblas_dtrsv (const int N, gslgmp_matrix* A, gslgmp_vector *X)
{
  mpfr_t tmp1,tmp2,tmp3,tmp4;
  mpfr_init2(tmp1,PRECBITS);
  mpfr_init2(tmp2,PRECBITS);
  mpfr_init2(tmp3,PRECBITS);
  mpfr_init2(tmp4,PRECBITS);



  const int nonunit = 1;
  int ix, jx;
  int i, j;

  if (N == 0)
    return;

  /* form  x := inv( A )*x */

    /* backsubstitution */
//We are not directly operating on the the memory but we are using the[] operator.

//    ix = OFFSET(N, incX) + incX * (N - 1);
      ix = OFFSET(N,1)+(N-1);

    if (nonunit) {
//      X[ix] = X[ix] / A[lda * (N - 1) + (N - 1)];
	mpfr_div(tmp1,X->data[ix],A->data[A->size2 * (N - 1) + (N - 1)],GMP_RNDN);
	mpfr_set(X->data[ix],tmp1,GMP_RNDN);

    }
//    ix -= incX;
    ix -= 1;
    for (i = N - 1; i > 0 && i--;) {
//      BASE tmp = X[ix];
	mpfr_set(tmp1,X->data[ix],GMP_RNDN);
//      jx = ix + incX;
      jx = ix + 1;
      for (j = i + 1; j < N; j++) {
//        const BASE Aij = A[lda * i + j];
        mpfr_set(tmp2,A->data[A->size2 * i + j],GMP_RNDN);
	mpfr_mul(tmp3,tmp2,X->data[jx],GMP_RNDN);
	mpfr_sub(tmp4,tmp1,tmp3,GMP_RNDN);
	mpfr_set(tmp1,tmp4,GMP_RNDN);
//        tmp -= Aij * X[jx];
//        jx += incX;
        jx += 1;
      }


      if (nonunit) {
//        X[ix] = tmp / A[lda * i + i];
        mpfr_div(X->data[ix],tmp1, A->data[A->size2 * i + i],GMP_RNDN);

      } else {
//        X[ix] = tmp;
        mpfr_set(X->data[ix],tmp1,GMP_RNDN);
      }
//      ix -= incX;
      ix -= 1;
    }

    mpfr_clear(tmp1);
    mpfr_clear(tmp2);
    mpfr_clear(tmp3);
    mpfr_clear(tmp4);

}

void 
gslgmp_blas_dtrsv (gslgmp_matrix * A, gslgmp_vector * X)
{
  const int M = A->size1;
  const int N = A->size2;

  if (M != N)
    {
      fprintf(stderr,"matrix must be square");
    }
  else if (N != X->size)
    {
      fprintf(stderr,"invalid length");
    }
  //CblasRowMajor, CblasUpper, CblasNoTrans, CblasNonUnit,
  gmpcblas_dtrsv ( N, A, X);
}



/* Solves the system A x = b in place using the QR factorisation,

 *  R x = Q^T b
 *
 * to obtain x. Based on SLATEC code. 
 */

void
gslgmp_linalg_householder_hv (mpfr_t tau, gslgmp_vector * v, gslgmp_vector * w)
{
  /* applies a householder transformation v to vector w */
  const int N = v->size;

  gslgmp_vector *v1,*w1;


  mpfr_t zero,d0,d1,d,w0,tmp,tmp2;
  mpfr_init2(zero,PRECBITS); 
  mpfr_init2(d,PRECBITS); 
  mpfr_init2(d1,PRECBITS); 
  mpfr_init2(d0,PRECBITS); 
  mpfr_init2(w0,PRECBITS); 
  mpfr_init2(tmp,PRECBITS); 
  mpfr_init2(tmp2,PRECBITS); 
  mpfr_set_d(zero,0.0,GMP_RNDN);


  if (mpfr_cmp(tau,zero) == 0){

	mpfr_clear(d);
	mpfr_clear(d0);
	mpfr_clear(d1);
	mpfr_clear(zero);
	mpfr_clear(w0);
	mpfr_clear(tmp);
	mpfr_clear(tmp2);

	return;}

  
    /* compute d = v'w */

//    double d0 = gsl_vector_get(w,0);
    gslgmp_vector_get(w,0,d0);

//     gsl_vector_const_view v1 = gsl_vector_const_subvector(v, 1, N-1);
//     gsl_vector_view w1 = gsl_vector_subvector(w, 1, N-1);
    v1=gslgmp_vector_calloc(N-1);
    w1=gslgmp_vector_calloc(N-1);

    gslgmp_vector_part(v,1,N-1,v1);
    gslgmp_vector_part(w,1,N-1,w1);

//    gsl_blas_ddot (&v1.vector, &w1.vector, &d1);
    gslgmp_ddot (v1, w1, d1);
    
//    d = d0 + d1;
    mpfr_add(d,d0,d1,GMP_RNDN);

    /* compute w = w - tau (v) (v'w) */
  
    
//      double w0 = gsl_vector_get (w,0);
      gslgmp_vector_get (w,0,w0);
      mpfr_mul(tmp,tau,d,GMP_RNDN);
      mpfr_sub(tmp2,w0,tmp,GMP_RNDN);
//      gsl_vector_set (w, 0, w0 - tau * d);
      gslgmp_vector_set (w, 0, tmp2);
    
    //y  <--  alpha*x + y
//    gsl_blas_daxpy (-tau * d, &v1.vector, &w1.vector);
   for(int k=0;k<v1->size;k++){

	mpfr_mul(tmp,tau,d,GMP_RNDN);
	mpfr_neg(tmp2,tmp,GMP_RNDN);
	mpfr_mul(tmp,v1->data[k],tmp2,GMP_RNDN); //CHeCK
	mpfr_add(tmp2,tmp,w1->data[k],GMP_RNDN);
	mpfr_set(w1->data[k],tmp2,GMP_RNDN);
   }




    gslgmp_vector_reverse_part(v,1,N-1,v1);
    gslgmp_vector_reverse_part(w,1,N-1,w1);


  

  mpfr_clear(d);
  mpfr_clear(d0);
  mpfr_clear(d1);
  mpfr_clear(zero);
  mpfr_clear(w0);
  mpfr_clear(tmp);
  mpfr_clear(tmp2);

  gslgmp_vector_free(v1);
  gslgmp_vector_free(w1);


}


void
gslgmp_linalg_QR_QTvec (gslgmp_matrix * QR, gslgmp_vector * tau, gslgmp_vector * v)
{
  const int M = QR->size1;
  const int N = QR->size2;

  gslgmp_vector *c,*h,*w;
  mpfr_t ti;
  mpfr_init2(ti,PRECBITS);

  if (tau->size != GSL_MIN (M, N))
    {
      fprintf(stderr,"size of tau must be MIN(M,N)");
    }
  else if (v->size != M)
    {
      fprintf(stderr,"vector size must be N");
    }
  else
    {
      int i;

      /* compute Q^T v */

      for (i = 0; i < GSL_MIN (M, N); i++)
        {
//           gsl_vector_const_view c = gsl_matrix_const_column (QR, i);
//           gsl_vector_const_view h = gsl_vector_const_subvector (&(c.vector), i, M - i);
//           gsl_vector_view w = gsl_vector_subvector (v, i, M - i);
	     	c=gslgmp_vector_calloc(QR->size1);
		h=gslgmp_vector_calloc(M-i);
		w=gslgmp_vector_calloc(M-i);

		gslgmp_matrix_column(QR,i,c);
		gslgmp_vector_part(c,i,M-i,h);
		gslgmp_vector_part(v,i,M-i,w);

//          double ti = gsl_vector_get (tau, i);
          gslgmp_vector_get (tau, i,ti);
//fprintf(stderr,"DEBUGG gmp :%d %g\n",i,mpfr_get_d(ti,GMP_RNDN));


	//ACHTUNG
          gslgmp_linalg_householder_hv (ti, h, w);

	  gslgmp_vector_reverse_part(c,i,M-i,h);
	  gslgmp_vector_reverse_part(v,i,M-i,w);
	  gslgmp_matrix_reverse_column(QR,i,c);

	gslgmp_vector_free(c);
	gslgmp_vector_free(h);
	gslgmp_vector_free(w);


        }
    }
  mpfr_clear(ti);
}

void
gslgmp_linalg_QR_svx (gslgmp_matrix * QR, gslgmp_vector * tau, gslgmp_vector * x)
{

  if (QR->size1 != QR->size2)
    {
      fprintf(stderr,"QR matrix must be square");
    }
  else if (QR->size1 != x->size)
    {
      fprintf(stderr,"matrix size must match x/rhs size");
    }
  else
    {
      /* compute rhs = Q^T b */
// 	fprintf(stderr,"QTvecbefore\n");
// 	gmpPMat(QR);
// 	gmpPVec(tau);
// 	gmpPVec(x);

     gslgmp_linalg_QR_QTvec (QR, tau, x);

// 	fprintf(stderr,"QTvecbafter\n");
// 	gmpPMat(QR);
// 	gmpPVec(tau);
// 	gmpPVec(x);
// 	fprintf(stderr,"To solver\n");


      /* Solve R x = rhs, storing x in-place */

//      gsl_blas_dtrsv (CblasUpper, CblasNoTrans, CblasNonUnit, QR, x);
      gslgmp_blas_dtrsv (QR, x);

	
// 	fprintf(stderr,"From solver\n");
// 	gmpPVec(x);
// 	gmpPMat(QR);


    }
}


void
gslgmp_linalg_QR_solve (gslgmp_matrix * QR, gslgmp_vector * tau, gslgmp_vector * b, gslgmp_vector * x)
{
  if (QR->size1 != QR->size2)
    {
      fprintf(stderr, "QR matrix must be square");exit(-1);
    }
  else if (QR->size1 != b->size)
    {
      fprintf(stderr,"matrix size must match b size");exit(-1);
    }
  else if (QR->size2 != x->size)
    {
      fprintf(stderr,"matrix size must match solution size");exit(-1);
    }
  else
    {
      /* Copy x <- b */

//      gsl_vector_memcpy (x, b);

 	if(b->size!=x->size){fprintf(stderr,"sth weired with solving algorithm\n"); exit(-1);}

	for(int k=0;k<b->size;k++){
	mpfr_set(x->data[k],b->data[k],GMP_RNDN);
	}

// 	gmpPVec(x);
// 	while(getchar()==0);


      /* Solve for x */

//      gsl_linalg_QR_svx (QR, tau, x);
      gslgmp_linalg_QR_svx (QR, tau, x);

      
    }
}


void gmpQRSolve(gslgmp_matrix* A, gslgmp_vector* b, gslgmp_vector* x)
{

	gslgmp_vector* btmp;
	btmp=gslgmp_vector_calloc(b->size);
	for(int i=0;i<b->size;i++){
	mpfr_set(btmp->data[i],b->data[i],GMP_RNDN);
	}


       	gslgmp_matrix* tmp;
	tmp=gslgmp_matrix_calloc(A->size1,A->size2);

	for(int i=0;i<A->size1*A->size2;i++){
	mpfr_set(tmp->data[i],A->data[i],GMP_RNDN);
	}
	
	gslgmp_vector* tau;
	tau=gslgmp_vector_calloc(A->size1);


	gslgmp_linalg_QR_decomp (tmp, tau);

/*	gmpPMat(tmp);
	gmpPVec(tau);*/
// 	while(getchar()==0);

	gslgmp_linalg_QR_solve (tmp, tau, btmp, x);


	gslgmp_matrix_free(tmp);
	gslgmp_vector_free(tau);
	gslgmp_vector_free(btmp);


}

void gmpQRInvert(gslgmp_matrix* in, gslgmp_matrix* out)
{

	if(in->size1!=in->size2){ fprintf(stderr,"INVERT OF NON NxN matrix exiting\n");exit(-1);}

	mpfr_t tmp1;
	mpfr_init2(tmp1,PRECBITS);

       	gslgmp_matrix* tmp;
	tmp=gslgmp_matrix_calloc(in->size1,in->size2);

       	gslgmp_matrix* tmpn2;
	gslgmp_vector* b;
	gslgmp_vector* x;

	b=gslgmp_vector_calloc(in->size1);
	x=gslgmp_vector_calloc(in->size1);
	tmpn2=gslgmp_matrix_calloc(in->size1,in->size2);

	gslgmp_vector* tau;
	gslgmp_vector* bt;
	tau=gslgmp_vector_calloc(in->size1);
	bt=gslgmp_vector_calloc(in->size1);

	gslgmp_matrix_memcpy(tmp,in); //(tmp<-in)
	gslgmp_linalg_QR_decomp (tmp, tau);

	for(int k=0;k<in->size1;k++){

		for(int j=0;j<in->size1;j++){
			gslgmp_vector_set_d(b,j,((j==k)?1.:0.));
		}
	
		gslgmp_linalg_QR_solve (tmp, tau, b, x);
	
		for(int j=0;j<in->size1;j++){
			gslgmp_vector_get(x,j,tmp1);
			gslgmp_matrix_set(out,j,k,tmp1);
		}

	}

	gslgmp_vector_free(b);
	gslgmp_vector_free(x);
	gslgmp_matrix_free(tmpn2);

	gslgmp_vector_free(bt);
	gslgmp_vector_free(tau);
	gslgmp_matrix_free(tmp);
	mpfr_clear(tmp1);

}


//**********************************************************************************
//  DATA and MOCK is still handled as double: Print routines
//**********************************************************************************

void DblPMat(double *in, int N1, int N2)
{
	fprintf(stderr,"\n\n{");
	for(int i=0;i<N1;i++){
	  fprintf(stderr,"{");
	  for(int j=0;j<N2;j++){
	    if(j==(N2-1)){fprintf(stderr,"%.7e\t",in[i+N1*j]);}else{fprintf(stderr,"%.7e,\t",in[i+N1*j]);}
	  } fprintf(stderr,"},\n");
	}	fprintf(stderr,"}\n\n");

}

void DblPMatT(double *in, int N1, int N2)
{
	fprintf(stderr,"\n\n{");
	for(int i=0;i<N2;i++){
	  fprintf(stderr,"{");
	  for(int j=0;j<N1;j++){
	    if(j==(N1-1)){fprintf(stderr,"%.4e\t",in[i+N2*j]);}else{fprintf(stderr,"%.4e,\t",in[i+N2*j]);}
	  } fprintf(stderr,"},\n");
	}	fprintf(stderr,"}\n\n");

}

void DblPVec(double* in,int N)
{
	fprintf(stderr,"\n\n{");
	  for(int j=0;j<N;j++){
	    fprintf(stderr,"%.4e,\t",in[j]);
	  } fprintf(stderr,"}\n\n");
}


// void PMat(gsl_matrix* in)
// {
// 	fprintf(stderr,"\n\n{");
// 	for(int i=0;i<in->size1;i++){
// 	  fprintf(stderr,"{");
// 	  for(int j=0;j<in->size2;j++){
// 	    if(j==(in->size2-1)){fprintf(stderr,"%.4e\t",gsl_matrix_get(in,i,j));}else{fprintf(stderr,"%.4e,\t",gsl_matrix_get(in,i,j));}
// 	  } fprintf(stderr,"},\n");
// 	}	fprintf(stderr,"}\n\n");
// 
// }
// 
// void PVec(gsl_vector* in)
// {
// 	fprintf(stderr,"\n\n{");
// 	  for(int j=0;j<in->size;j++){
// 	    fprintf(stderr,"%.4e,\t",gsl_vector_get(in,j));
// 	  } fprintf(stderr,"}\n\n");
// }


void PArr(const double* in,int N)
{
	fprintf(stderr,"\n\n{");
	  for(int j=0;j<N;j++){
	    fprintf(stderr,"%.4e,\t",in[j]);
	  } fprintf(stderr,"}\n\n");
}


//**********************************************************************************
//  Correlations and Alpha weights are the part that need Eigenvalue decomposition,
//**********************************************************************************

void gslgmp_vector_scale(gslgmp_vector* in, mpfr_t factor)
{
	mpfr_t tmp1;
	mpfr_init2(tmp1,PRECBITS);

	for(int i=0; i< in->size;i++){

	mpfr_set(tmp1,in->data[i],GMP_RNDN);
	mpfr_mul(in->data[i],tmp1,factor,GMP_RNDN);

	}

 	mpfr_clear(tmp1);

}

void gslgmp_daxpy(mpfr_t alpha, gslgmp_vector* x, gslgmp_vector* y)
{
	if(x->size!=y->size) {fprintf(stderr,"Vector size does not match!!"); exit(-1);}

	mpfr_t tmp1;
	mpfr_init2(tmp1,PRECBITS);
	mpfr_t tmp2;
	mpfr_init2(tmp2,PRECBITS);

	for(int i=0; i< x->size;i++){

	mpfr_set(tmp1,x->data[i],GMP_RNDN);
	mpfr_mul(tmp2,tmp1,alpha,GMP_RNDN);
	mpfr_set(tmp1,y->data[i],GMP_RNDN);
	mpfr_add(y->data[i],tmp1,tmp2,GMP_RNDN);

	}

 	mpfr_clear(tmp1);
 	mpfr_clear(tmp2);

}


void gslgmp_dsyr2(mpfr_t alpha, gslgmp_vector* x, gslgmp_vector* y,gslgmp_matrix* A)
{
	if(x->size!=y->size || x->size!=A->size1 || x->size!=A->size2 ) {fprintf(stderr,"Vector Matrix size does not match!!"); exit(-1);}

	gslgmp_matrix *tmp;
	tmp=gslgmp_matrix_calloc(A->size1,A->size2);

	mpfr_t tmp1;
	mpfr_init2(tmp1,PRECBITS);
	mpfr_t tmp2;
	mpfr_init2(tmp2,PRECBITS);
	mpfr_t tmp3;
	mpfr_init2(tmp3,PRECBITS);

//Build the first factor alpha x y^T
	for(int i=0; i< A->size1;i++){
	  for(int j=0; j< A->size2;j++){
	   mpfr_mul(tmp1,x->data[i],y->data[j],GMP_RNDN);
	   mpfr_mul(tmp2,tmp1,alpha,GMP_RNDN);
	   gslgmp_matrix_set(tmp,i,j,tmp2);
	}}

//Second factor is just transpose since symmetric, add to A

	for(int i=0; i< A->size1;i++){
	  for(int j=0; j< A->size2;j++){
	   gslgmp_matrix_get(tmp,i,j,tmp1);
	   gslgmp_matrix_get(tmp,j,i,tmp2);
	   mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
	   gslgmp_matrix_get(A,i,j,tmp1);
	   mpfr_add(tmp2,tmp1,tmp3,GMP_RNDN);
	   gslgmp_matrix_set(A,i,j,tmp2);
	}}


 	mpfr_clear(tmp1);
 	mpfr_clear(tmp2);
 	mpfr_clear(tmp3);
	gslgmp_matrix_free(tmp);
}


void gslgmp_dsyr2_lower(mpfr_t alpha, gslgmp_vector* x, gslgmp_vector* y,gslgmp_matrix* A)
{
	if(x->size!=y->size || x->size!=A->size1 || x->size!=A->size2 ) {fprintf(stderr,"Vector Matrix size does not match!!"); exit(-1);}

	gslgmp_matrix *tmp;
	tmp=gslgmp_matrix_calloc(A->size1,A->size2);

	mpfr_t tmp1;
	mpfr_init2(tmp1,PRECBITS);
	mpfr_t tmp2;
	mpfr_init2(tmp2,PRECBITS);
	mpfr_t tmp3;
	mpfr_init2(tmp3,PRECBITS);

//Build the first factor alpha x y^T
	for(int i=0; i< A->size1;i++){
	  for(int j=0; j< A->size2;j++){
	   mpfr_mul(tmp1,x->data[i],y->data[j],GMP_RNDN);
	   mpfr_mul(tmp2,tmp1,alpha,GMP_RNDN);
	   gslgmp_matrix_set(tmp,i,j,tmp2);
	}}

//Second factor is just transpose since symmetric, add to A lower half

	for(int i=0; i< A->size1;i++){
	  for(int j=0; j<=i;j++){
	   gslgmp_matrix_get(tmp,i,j,tmp1);
	   gslgmp_matrix_get(tmp,j,i,tmp2);
	   mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
	   gslgmp_matrix_get(A,i,j,tmp1);
	   mpfr_add(tmp2,tmp1,tmp3,GMP_RNDN);
	   gslgmp_matrix_set(A,i,j,tmp2);
	}}

 	mpfr_clear(tmp1);
 	mpfr_clear(tmp2);
 	mpfr_clear(tmp3);
	gslgmp_matrix_free(tmp);
}

void
gslgmp_linalg_symmtd_decomp (gslgmp_matrix * A, gslgmp_vector * tau)
{
  if (A->size1 != A->size2)
    {
      fprintf(stderr,"symmetric tridiagonal decomposition requires square matrix");
    }
  else if (tau->size + 1 != A->size1)
    {
      fprintf(stderr,"size of tau must be (matrix size - 1)");
    }
  else
    {
      const int N = A->size1;
      int i;

  	mpfr_t tau_i,ei,xv,alpha,zero,minusone,tmp1;
	mpfr_init2(tau_i,PRECBITS);
	mpfr_init2(ei,PRECBITS);
	mpfr_init2(xv,PRECBITS);
	mpfr_init2(tmp1,PRECBITS);
	mpfr_init2(alpha,PRECBITS);
	mpfr_init2(zero,PRECBITS);
	mpfr_set_d(zero,0.0,GMP_RNDN);
	mpfr_init2(minusone,PRECBITS);
	mpfr_set_d(minusone,-1.0,GMP_RNDN);
	gslgmp_vector* view_c;
	gslgmp_vector* view_v;
	gslgmp_vector* view_x;
	gslgmp_matrix* view_m;
	

      for (i = 0 ; i < N - 2; i++)
        {
	view_c=gslgmp_vector_calloc(A->size1);
	view_v=gslgmp_vector_calloc(N-(i+1));

	gslgmp_matrix_column(A,i,view_c);
	gslgmp_vector_part(view_c,i+1,N-(i+1),view_v);
	gslgmp_linalg_householder_transform(view_v,tau_i);

         // gsl_vector_view c = gsl_matrix_column (A, i);
         // gsl_vector_view v = gsl_vector_subvector (&c.vector, i + 1, N - (i + 1));
         // double tau_i = gsl_linalg_householder_transform (&v.vector);
          
// 	gmpPVec(view_c);
// 	gmpPVec(view_v);


          // Apply the transformation H^T A H to the remaining columns 

          if (mpfr_cmp(tau_i,zero) != 0) 
            {
	      	view_m=gslgmp_matrix_calloc(N - (i+1),N - (i+1));
		gslgmp_submatrix(A,i+1,i+1,N - (i+1), N - (i+1),view_m);

              //gsl_matrix_view m = gsl_matrix_submatrix (A, i + 1, i + 1, 
              //                                          N - (i+1), N - (i+1));
              //double ei = gsl_vector_get(&v.vector, 0);
              gslgmp_vector_get(view_v, 0,ei);

              //gsl_vector_view x = gsl_vector_subvector (tau, i, N-(i+1));
              //gsl_vector_set (&v.vector, 0, 1.0);

		view_x=gslgmp_vector_calloc(N-(i+1));
		gslgmp_vector_part(tau,i,(N-(i+1)),view_x);
		gslgmp_vector_set_d(view_v,0,1.0);


              
              // x = tau * A * v 
              //gsl_blas_dsymv (CblasLower, tau_i, &m.matrix, &v.vector, 0.0, &x.vector);
	      gmp_multmv_lower(view_m,view_v,view_x);
	      gslgmp_vector_scale(view_x,tau_i);

// // 	 	gmpPVec(view_x);
// // 	 	gmpPVec(view_v);
// // 		gmpPMat(view_m);

              // w = x - (1/2) tau * (x' * v) * v  
              {
                //double xv, alpha;
                //gsl_blas_ddot(&x.vector, &v.vector, &xv);
		gslgmp_ddot(view_x,view_v,xv);

                //alpha = - (tau_i / 2.0) * xv;
		mpfr_div_ui(tmp1,tau_i,2,GMP_RNDN);
		mpfr_neg(tmp1,tmp1,GMP_RNDN);
		mpfr_mul(alpha,tmp1,xv,GMP_RNDN);

                //gsl_blas_daxpy(alpha, &v.vector, &x.vector);
		gslgmp_daxpy(alpha,view_v,view_x);


              }

              // apply the transformation A = A - v w' - w v' 
              //gsl_blas_dsyr2(CblasLower, -1.0, &v.vector, &x.vector, &m.matrix);
              gslgmp_dsyr2_lower(minusone, view_v, view_x, view_m);

              gslgmp_vector_set (view_v, 0, ei);

// 		fprintf(stderr,"PREC DEB %g\n",mpfr_get_d(tau_i,GMP_RNDN));
// 	 	gmpPVec(view_x);
// 	 	gmpPVec(view_v);
// 		gmpPMat(view_m);



      	      gslgmp_vector_reverse_part(tau,i,(N-(i+1)),view_x);
  	      gslgmp_reverse_submatrix(A,i+1,i+1,N - (i+1), N - (i+1),view_m);
	      gslgmp_vector_reverse_part(view_c,i+1,N-(i+1),view_v);
	      gslgmp_matrix_reverse_column(A,i,view_c);
	
		gslgmp_vector_free(view_c);
		gslgmp_vector_free(view_v);
		gslgmp_vector_free(view_x);
		gslgmp_matrix_free(view_m);



            }
          
          gslgmp_vector_set (tau, i, tau_i);
        }
      
      //return GSL_SUCCESS;

  	mpfr_clear(tau_i);
  	mpfr_clear(ei);
  	mpfr_clear(xv);
 	mpfr_clear(tmp1);
  	mpfr_clear(alpha);
  	mpfr_clear(zero);
  	mpfr_clear(minusone);


    }


}  



void gslgmp_linalg_symmtd_unpack (gslgmp_matrix * A, 
                          gslgmp_vector * tau,
                          gslgmp_matrix * Q, 
                          gslgmp_vector * diag, 
                          gslgmp_vector * sdiag)
{
  if (A->size1 !=  A->size2)
    {
      fprintf(stderr,"matrix A must be square");
    }
  else if (tau->size + 1 != A->size1)
    {
      fprintf(stderr,"size of tau must be (matrix size - 1)");
    }
  else if (Q->size1 != A->size1 || Q->size2 != A->size1)
    {
      fprintf(stderr,"size of Q must match size of A");
    }
  else if (diag->size != A->size1)
    {
      fprintf(stderr,"size of diagonal must match size of A");
    }
  else if (sdiag->size + 1 != A->size1)
    {
      fprintf(stderr,"size of subdiagonal must be (matrix size - 1)");
    }
  else
    {
      const int N = A->size1;

      int i;

     	mpfr_t ti,Aii,Aji;
	mpfr_init2(ti,PRECBITS);
	mpfr_init2(Aii,PRECBITS);
	mpfr_init2(Aji,PRECBITS);

	gslgmp_vector* view_c;
	gslgmp_vector* view_h;
	gslgmp_matrix* view_m;



      // Initialize Q to the identity 

      gslgmp_matrix_set_identity (Q);

      for (i = N - 2; i > 0 && i--;)
        {
          //gsl_vector_const_view c = gsl_matrix_const_column (A, i);
          //gsl_vector_const_view h = gsl_vector_const_subvector (&c.vector, i + 1, N - (i+1));
	  view_c=gslgmp_vector_calloc(A->size1);
	  view_h=gslgmp_vector_calloc(N-(i+1));
	  gslgmp_matrix_column(A,i,view_c);
	  gslgmp_vector_part(view_c,i+1,N-(i+1),view_h);

          gslgmp_vector_get (tau, i,ti);

          //gsl_matrix_view m = gsl_matrix_submatrix (Q, i + 1, i + 1, N-(i+1), N-(i+1));
	  view_m=gslgmp_matrix_calloc(N-(i+1),N-(i+1));
	  gslgmp_submatrix(Q,i+1,i+1,N-(i+1), N-(i+1),view_m);


          gslgmp_linalg_householder_hm (ti, view_h, view_m);

	  gslgmp_vector_reverse_part(view_c,i+1,N-(i+1),view_h);	 
	  gslgmp_matrix_reverse_column(A,i,view_c);
	  gslgmp_reverse_submatrix(Q,i+1,i+1,N-(i+1), N-(i+1),view_m);

	  gslgmp_vector_free(view_c);
	  gslgmp_vector_free(view_h);
	  gslgmp_matrix_free(view_m);


        }

      // Copy diagonal into diag

      for (i = 0; i < N; i++)
        {
          gslgmp_matrix_get (A, i, i,Aii);
          gslgmp_vector_set (diag, i, Aii);
        }

      // Copy subdiagonal into sd 

      for (i = 0; i < N - 1; i++)
        {
          gslgmp_matrix_get (A, i+1, i,Aji);
          gslgmp_vector_set (sdiag, i, Aji);
        }

	mpfr_clear(ti);
	mpfr_clear(Aii);
	mpfr_clear(Aji);


    }
}


void gslgmp_linalg_symmtd_unpack_T (gslgmp_matrix * A, 
                            gslgmp_vector * diag, 
                            gslgmp_vector * sdiag)
{
  if (A->size1 !=  A->size2)
    {
      fprintf(stderr,"matrix A must be square");
    }
  else if (diag->size != A->size1)
    {
      fprintf(stderr,"size of diagonal must match size of A");
    }
  else if (sdiag->size + 1 != A->size1)
    {
      fprintf(stderr,"size of subdiagonal must be (matrix size - 1)");
    }
  else
    {
      const int N = A->size1;

      int i;

      mpfr_t Aii, Aji;

      mpfr_init2(Aii,PRECBITS);
      mpfr_init2(Aji,PRECBITS);

      /* Copy diagonal into diag */

      for (i = 0; i < N; i++)
        {
//           double Aii = gsl_matrix_get (A, i, i);
//           gsl_vector_set (diag, i, Aii);

          gslgmp_matrix_get (A, i, i,Aii);
          gslgmp_vector_set (diag, i, Aii);

        }

      /* Copy subdiagonal into sdiag */

      for (i = 0; i < N - 1; i++)
        {
//           double Aij = gsl_matrix_get (A, i+1, i);
//           gsl_vector_set (sdiag, i, Aij);

          gslgmp_matrix_get (A, i+1, i,Aji);
          gslgmp_vector_set (sdiag, i, Aji);
        }

	mpfr_clear(Aii);
	mpfr_clear(Aji);

    }
}



void gmpcreate_givens (mpfr_t a, mpfr_t b, mpfr_t c, mpfr_t s)
{
  mpfr_t tmp1,tmp2,tmp3;
  mpfr_init2(tmp1,PRECBITS);
  mpfr_init2(tmp2,PRECBITS);
  mpfr_init2(tmp3,PRECBITS);


  if (mpfr_cmp_ui(b,0) == 0)
    {
      mpfr_set_d(c,1.,GMP_RNDN);
      mpfr_set_d(s,0.,GMP_RNDN);
    }
  else if (mpfr_cmpabs(b,a) >0 )
    {
	mpfr_div(tmp1,a,b,GMP_RNDN);
	mpfr_neg(tmp1,tmp1,GMP_RNDN);

	mpfr_mul(tmp2,tmp1,tmp1,GMP_RNDN);
	mpfr_add_ui(tmp3,tmp2,1,GMP_RNDN);
	mpfr_sqrt(tmp2,tmp3,GMP_RNDN);
	mpfr_ui_div(tmp3,1,tmp2,GMP_RNDN);

	mpfr_set(s,tmp3,GMP_RNDN);
	mpfr_mul(c,tmp3,tmp1,GMP_RNDN);

//      double t = -a / b;
//      double s1 = 1.0 / sqrt (1 + t * t);
//      *s = s1;
//      *c = s1 * t;

    }
  else
    {
	mpfr_div(tmp1,b,a,GMP_RNDN);
	mpfr_neg(tmp1,tmp1,GMP_RNDN);

	mpfr_mul(tmp2,tmp1,tmp1,GMP_RNDN);
	mpfr_add_ui(tmp3,tmp2,1,GMP_RNDN);
	mpfr_sqrt(tmp2,tmp3,GMP_RNDN);
	mpfr_ui_div(tmp3,1,tmp2,GMP_RNDN);

	mpfr_set(c,tmp3,GMP_RNDN);
	mpfr_mul(s,tmp3,tmp1,GMP_RNDN);

//     double t = -b / a;
//      double c1 = 1.0 / sqrt (1 + t * t);
//      *c = c1;
//      *s = c1 * t;
    }

    mpfr_clear(tmp1);
    mpfr_clear(tmp2);
    mpfr_clear(tmp3);


}

 //eigenval routines work until here

void gmptrailing_eigenvalue (const int n, gslgmp_vector* d, gslgmp_vector* sd, mpfr_t mu)
{
//   double ta = d[n - 2];
//   double tb = d[n - 1];
//   double tab = sd[n - 2];

//   double dt = (ta - tb) / 2.0;

//  mpfr_set(ta,d->data[n-2],GMP_RNDN);
//  mpfr_set(tb,d->data[n-1],GMP_RNDN);
//  mpfr_set(tab,sd->data[n-2],GMP_RNDN);

  mpfr_t tmp1,tmp2,dt;
  mpfr_init2(tmp1,PRECBITS);
  mpfr_init2(tmp2,PRECBITS);
  mpfr_init2(dt,PRECBITS);



  mpfr_sub(tmp1,d->data[n-2],d->data[n-1],GMP_RNDN);
  mpfr_div_ui(dt,tmp1,2,GMP_RNDN);


  if (mpfr_sgn(dt) > 0)
    {
     mpfr_hypot(tmp1,dt,sd->data[n-2],GMP_RNDN);
     mpfr_add(tmp2,dt,tmp1,GMP_RNDN);
     mpfr_div(tmp1,sd->data[n-2],tmp2,GMP_RNDN);
     mpfr_mul(tmp2,tmp1,sd->data[n-2],GMP_RNDN);
     mpfr_sub(mu,d->data[n-1],tmp2,GMP_RNDN);
     // mu = tb - tab * (tab / (dt + hypot (dt, tab)));
    }
  else if (mpfr_sgn(dt) == 0) 
    {
	gmpabs(tmp1,sd->data[n-2]);
	mpfr_sub(mu,d->data[n-1],tmp1,GMP_RNDN);

      //mu = tb - fabs(tab);
    }
  else
    {
     mpfr_hypot(tmp1,dt,sd->data[n-2],GMP_RNDN);
     mpfr_sub(tmp2,tmp1,dt,GMP_RNDN);
     mpfr_div(tmp1,sd->data[n-2],tmp2,GMP_RNDN);
     mpfr_mul(tmp2,tmp1,sd->data[n-2],GMP_RNDN);
     mpfr_add(mu,d->data[n-1],tmp2,GMP_RNDN);
//     mu = tb + tab * (tab / ((-dt) + hypot (dt, tab)));
    }

    mpfr_clear(dt);
    mpfr_clear(tmp1);
    mpfr_clear(tmp2);


}


static void
gmpchop_small_elements (const int N, gslgmp_vector *d, gslgmp_vector* sd)
{
  mpfr_t d_i,sd_i,d_ip1,myprec,tmp1,tmp2,tmp3;
  mpfr_init2(tmp1,PRECBITS);
  mpfr_init2(tmp2,PRECBITS);
  mpfr_init2(tmp3,PRECBITS);
  mpfr_init2(d_i,PRECBITS);
  mpfr_init2(sd_i,PRECBITS);
  mpfr_init2(d_ip1,PRECBITS);
  mpfr_init2(myprec,PRECBITS);
  mpfr_ui_pow_ui(myprec,2,PRECBITS,GMP_RNDN);
  mpfr_ui_div(myprec,1,myprec,GMP_RNDN);
  //mpfr_set_d(myprec,(pow(2,-PRECBITS)),GMP_RNDN);

  gslgmp_vector_get(d,0,d_i);

  int i;

  for (i = 0; i < N - 1; i++)
    {
  	gslgmp_vector_get(sd,i,sd_i);
  	gslgmp_vector_get(d,i+1,d_ip1);
//       double sd_i = sd[i];
//       double d_ip1 = d[i + 1];

	gmpabs(d_ip1,tmp1);
	gmpabs(d_i,tmp2);
	mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
	mpfr_mul(tmp1,tmp3,myprec,GMP_RNDN);
	gmpabs(sd_i,tmp2);

      //if (fabs (sd_i) < GSL_DBL_EPSILON * (fabs (d_i) + fabs (d_ip1)))
	if (mpfr_cmp(tmp2,tmp1)<0)
        {
          gslgmp_vector_set_d(sd,i,0.); //sd[i] = 0.0;
        }
      mpfr_set(d_i,d_ip1,GMP_RNDN);
    }

 mpfr_clear(d_i);
 mpfr_clear(sd_i);
 mpfr_clear(d_ip1);
 mpfr_clear(myprec);
 mpfr_clear(tmp1);
 mpfr_clear(tmp2);
 mpfr_clear(tmp3);
}


static void
gmpqrstep (const int n, gslgmp_vector *d, gslgmp_vector* sd, gslgmp_vector* gc, gslgmp_vector *gs)
{
  mpfr_t x, z,mu,c,s,tmp1,tmp2,tmp3;
  mpfr_t ak, bk, zk, ap, bp, aq, bq,aq1,ap1,bp1,bk1,zp1,bq1;
  int k;

  mpfr_init2(x,PRECBITS);
  mpfr_init2(z,PRECBITS);
  mpfr_init2(mu,PRECBITS);
  mpfr_init2(ak,PRECBITS);
  mpfr_init2(bk,PRECBITS);
  mpfr_init2(zk,PRECBITS);
  mpfr_init2(ap,PRECBITS);
  mpfr_init2(bq,PRECBITS);
  mpfr_init2(aq,PRECBITS);
  mpfr_init2(bp,PRECBITS);
  mpfr_init2(c,PRECBITS);
  mpfr_init2(s,PRECBITS);
  mpfr_init2(aq1,PRECBITS);
  mpfr_init2(ap1,PRECBITS);
  mpfr_init2(bp1,PRECBITS);
  mpfr_init2(bk1,PRECBITS);
  mpfr_init2(zp1,PRECBITS);
  mpfr_init2(bq1,PRECBITS);
 mpfr_init2(tmp1,PRECBITS);
 mpfr_init2(tmp2,PRECBITS);
 mpfr_init2(tmp3,PRECBITS);


//fprintf(stderr, "DEBUG gmpqr 1\n");
  gmptrailing_eigenvalue (n, d, sd,mu);
//fprintf(stderr, "DEBUG gmpqr 2\n");

  //x = d[0] - mu;
  mpfr_sub(x,d->data[0],mu,GMP_RNDN);
  //z = sd[0];
  mpfr_set(z,sd->data[0],GMP_RNDN);

  mpfr_set_d(ak,0.,GMP_RNDN);
  mpfr_set_d(bk,0.,GMP_RNDN);
  mpfr_set_d(zk,0.,GMP_RNDN);
//fprintf(stderr, "DEBUG gmpqr 3\n");

  //ap = d[0];
  mpfr_set(ap,d->data[0],GMP_RNDN);
  //bp = sd[0];
  mpfr_set(bp,sd->data[0],GMP_RNDN);

  //aq = d[1];
  mpfr_set(aq,d->data[1],GMP_RNDN);
//fprintf(stderr, "DEBUG gmpqr 4\n");

  if (n == 2)
    {
//fprintf(stderr, "DEBUG gmpqr 5\n");
       gmpcreate_givens (x, z, c, s);
//fprintf(stderr, "DEBUG gmpqr 6\n");

      if (gc != NULL)
        mpfr_set(gc->data[0],c,GMP_RNDN); 
      if (gs != NULL)
        mpfr_set(gs->data[0],s,GMP_RNDN);

      {
//        double ap1 = c * (c * ap - s * bp) + s * (s * aq - c * bp);
//fprintf(stderr, "DEBUG gmpqr 7\n");
	mpfr_mul(tmp1,c,bp,GMP_RNDN);
	mpfr_mul(tmp2,s,aq,GMP_RNDN);
	mpfr_sub(tmp3,tmp2,tmp1,GMP_RNDN);
	mpfr_mul(ap1,s,tmp3,GMP_RNDN);

	mpfr_mul(tmp1,c,ap,GMP_RNDN);
	mpfr_mul(tmp2,s,bp,GMP_RNDN);
	mpfr_sub(tmp3,tmp1,tmp2,GMP_RNDN);
	mpfr_mul(tmp1,c,tmp3,GMP_RNDN);

	mpfr_set(tmp2,ap1,GMP_RNDN);
	mpfr_add(ap1,tmp2,tmp1,GMP_RNDN);
//fprintf(stderr, "DEBUG gmpqr 8\n");


//        double bp1 = c * (s * ap + c * bp) - s * (s * bp + c * aq);
	mpfr_mul(tmp1,c,aq,GMP_RNDN);
	mpfr_mul(tmp2,s,bp,GMP_RNDN);
	mpfr_add(tmp3,tmp2,tmp1,GMP_RNDN);
	mpfr_mul(bp1,s,tmp3,GMP_RNDN);

	mpfr_mul(tmp1,c,bp,GMP_RNDN);
	mpfr_mul(tmp2,s,ap,GMP_RNDN);
	mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
	mpfr_mul(tmp1,c,tmp3,GMP_RNDN);

	mpfr_set(tmp2,bp1,GMP_RNDN);
	mpfr_sub(bp1,tmp1,tmp2,GMP_RNDN);

 //       double aq1 = s * (s * ap + c * bp) + c * (s * bp + c * aq);

	mpfr_mul(tmp1,c,aq,GMP_RNDN);
	mpfr_mul(tmp2,s,bp,GMP_RNDN);
	mpfr_add(tmp3,tmp2,tmp1,GMP_RNDN);
	mpfr_mul(aq1,c,tmp3,GMP_RNDN);
//fprintf(stderr, "DEBUG gmpqr 9\n");

	mpfr_mul(tmp1,c,bp,GMP_RNDN);
	mpfr_mul(tmp2,s,ap,GMP_RNDN);
	mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
	mpfr_mul(tmp1,s,tmp3,GMP_RNDN);

	mpfr_set(tmp2,aq1,GMP_RNDN);
	mpfr_add(aq1,tmp1,tmp2,GMP_RNDN);

//fprintf(stderr, "DEBUG gmpqr 10\n");


	mpfr_set(ak,ap1,GMP_RNDN);
	mpfr_set(bk,bp1,GMP_RNDN);
	mpfr_set(ap,aq1,GMP_RNDN);

      }

      mpfr_set(d->data[0],ak,GMP_RNDN);
      mpfr_set(sd->data[0],bk,GMP_RNDN);
      mpfr_set(d->data[1],ap,GMP_RNDN);
//fprintf(stderr, "DEBUG gmpqr 11\n");

      mpfr_clear(x);
      mpfr_clear(z);
      mpfr_clear(mu);
      mpfr_clear(c);
      mpfr_clear(s);
      mpfr_clear(ak);
      mpfr_clear(bk);
      mpfr_clear(zk);
      mpfr_clear(ap);
      mpfr_clear(bp);
      mpfr_clear(aq);
      mpfr_clear(bq);
      mpfr_clear(aq1);
      mpfr_clear(ap1);
      mpfr_clear(bp1);
      mpfr_clear(bk1);
      mpfr_clear(zp1);
      mpfr_clear(bq1);
     mpfr_clear(tmp1);
     mpfr_clear(tmp2);
     mpfr_clear(tmp3);

      return;
    }

  mpfr_set(bq,sd->data[1],GMP_RNDN);

  for (k = 0; k < n - 1; k++)
    {
// fprintf(stderr, "DEBUG gmpqr 12\n");
     gmpcreate_givens (x, z, c, s);
//fprintf(stderr, "DEBUG gmpqr 13\n");

      // store Givens rotation
      if (gc != NULL)
        mpfr_set(gc->data[k],c,GMP_RNDN); 
      if (gs != NULL)
        mpfr_set(gs->data[k],s,GMP_RNDN);

      // compute G' T G 

      {
        //double bk1 = c * bk - s * zk;
//fprintf(stderr, "DEBUG gmpqr 14\n");

	mpfr_mul(tmp1,s,zk,GMP_RNDN);
	mpfr_mul(tmp2,c,bk,GMP_RNDN);
	mpfr_sub(bk1,tmp2,tmp1,GMP_RNDN);

//        double ap1 = c * (c * ap - s * bp) + s * (s * aq - c * bp);

	mpfr_mul(tmp1,c,bp,GMP_RNDN);
	mpfr_mul(tmp2,s,aq,GMP_RNDN);
	mpfr_sub(tmp3,tmp2,tmp1,GMP_RNDN);
	mpfr_mul(ap1,s,tmp3,GMP_RNDN);

	mpfr_mul(tmp1,c,ap,GMP_RNDN);
	mpfr_mul(tmp2,s,bp,GMP_RNDN);
	mpfr_sub(tmp3,tmp1,tmp2,GMP_RNDN);
	mpfr_mul(tmp1,c,tmp3,GMP_RNDN);

	mpfr_set(tmp2,ap1,GMP_RNDN);
	mpfr_add(ap1,tmp2,tmp1,GMP_RNDN);

//        double bp1 = c * (s * ap + c * bp) - s * (s * bp + c * aq);

	mpfr_mul(tmp1,c,aq,GMP_RNDN);
	mpfr_mul(tmp2,s,bp,GMP_RNDN);
	mpfr_add(tmp3,tmp2,tmp1,GMP_RNDN);
	mpfr_mul(bp1,s,tmp3,GMP_RNDN);

	mpfr_mul(tmp1,c,bp,GMP_RNDN);
	mpfr_mul(tmp2,s,ap,GMP_RNDN);
	mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
	mpfr_mul(tmp1,c,tmp3,GMP_RNDN);

	mpfr_set(tmp2,bp1,GMP_RNDN);
	mpfr_sub(bp1,tmp1,tmp2,GMP_RNDN);

//        double zp1 = -s * bq;

	mpfr_mul(tmp1,s,bq,GMP_RNDN);
	mpfr_neg(zp1,tmp1,GMP_RNDN);

//        double aq1 = s * (s * ap + c * bp) + c * (s * bp + c * aq);

	mpfr_mul(tmp1,c,aq,GMP_RNDN);
	mpfr_mul(tmp2,s,bp,GMP_RNDN);
	mpfr_add(tmp3,tmp2,tmp1,GMP_RNDN);
	mpfr_mul(aq1,c,tmp3,GMP_RNDN);

	mpfr_mul(tmp1,c,bp,GMP_RNDN);
	mpfr_mul(tmp2,s,ap,GMP_RNDN);
	mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
	mpfr_mul(tmp1,s,tmp3,GMP_RNDN);

	mpfr_set(tmp2,aq1,GMP_RNDN);
	mpfr_add(aq1,tmp1,tmp2,GMP_RNDN);

//        double bq1 = c * bq;

	mpfr_mul(bq1,c,bq,GMP_RNDN);


        mpfr_set(ak,ap1,GMP_RNDN);
        mpfr_set(bk,bp1,GMP_RNDN);
        mpfr_set(zk,zp1,GMP_RNDN);

        mpfr_set(ap,aq1,GMP_RNDN);
        mpfr_set(bp,bq1,GMP_RNDN);
//fprintf(stderr, "DEBUG gmpqr 15\n");

        if (k < n - 2)
          mpfr_set(aq,d->data[k + 2],GMP_RNDN);
        if (k < n - 3)
          mpfr_set(bq,sd->data[k + 2],GMP_RNDN);

        mpfr_set(d->data[k],ak,GMP_RNDN);

        if (k > 0)
          mpfr_set(sd->data[k - 1],bk1,GMP_RNDN);

        if (k < n - 2)
          mpfr_set(sd->data[k + 1],bp,GMP_RNDN);

        mpfr_set(x,bk,GMP_RNDN);
        mpfr_set(z,zk,GMP_RNDN);
      }
    }

  // k = n - 1 
  mpfr_set(d->data[k],ap,GMP_RNDN);
  mpfr_set(sd->data[k - 1],bk,GMP_RNDN);

      mpfr_clear(x);
      mpfr_clear(z);
      mpfr_clear(mu);
      mpfr_clear(c);
      mpfr_clear(s);
      mpfr_clear(ak);
      mpfr_clear(bk);
      mpfr_clear(zk);
      mpfr_clear(ap);
      mpfr_clear(bp);
      mpfr_clear(aq);
      mpfr_clear(bq);
      mpfr_clear(aq1);
      mpfr_clear(ap1);
      mpfr_clear(bp1);
      mpfr_clear(bk1);
      mpfr_clear(zp1);
      mpfr_clear(bq1);
     mpfr_clear(tmp1);
     mpfr_clear(tmp2);
     mpfr_clear(tmp3);



}



void gslgmp_eigen_symmv (gslgmp_matrix * A, gslgmp_vector * eval, gslgmp_matrix * evec)
{
  if (A->size1 != A->size2)
    {
      fprintf(stderr, "matrix must be square to compute eigenvalues");
    }
  else if (eval->size != A->size1)
    {
      fprintf(stderr,"eigenvalue vector must match matrix size");
    }
  else if (evec->size1 != A->size1 || evec->size2 != A->size1)
    {
      fprintf(stderr,"eigenvector matrix must match matrix size");
    }
  else
    {
 
     	const int N = A->size1;
      	int a, b;
     	mpfr_t A00,c,s,qki,qkj,zero,tmp1,tmp2,tmp3;
 	mpfr_init2(A00,PRECBITS);
 	mpfr_init2(c,PRECBITS);
 	mpfr_init2(s,PRECBITS);
 	mpfr_init2(qki,PRECBITS);
 	mpfr_init2(qkj,PRECBITS);
 	mpfr_init2(zero,PRECBITS);
	mpfr_init2(tmp1,PRECBITS);
	mpfr_init2(tmp2,PRECBITS);
	mpfr_init2(tmp3,PRECBITS);
	mpfr_set_d(zero,0.,GMP_RNDN);


      // handle special case 
      if (N == 1)
        {
          gslgmp_matrix_get (A, 0, 0,A00);
          gslgmp_vector_set (eval, 0, A00);
          gslgmp_matrix_set_d (evec, 0, 0, 1.0);
	  return;
         }

      // use sd as the temporary workspace for the decomposition when
      //   computing eigenvectors 

	gslgmp_vector* d;
	gslgmp_vector* sd;
	gslgmp_vector* tau;
	d=gslgmp_vector_calloc(N);
	sd=gslgmp_vector_calloc(N-1);
	tau=gslgmp_vector_calloc(N-1);

		gslgmp_vector *d_block;
		gslgmp_vector *sd_block;
		gslgmp_vector *gc;
		gslgmp_vector *gs;


      
       // gsl_vector_view d_vec = gsl_vector_view_array (d, N);
       // gsl_vector_view sd_vec = gsl_vector_view_array (sd, N - 1);
       // gsl_vector_view tau = gsl_vector_view_array (sd, N - 1);

	
        gslgmp_linalg_symmtd_decomp (A,tau);

	gslgmp_vector_reverse_part(sd,0,N-1,tau); //CHECK IF NEEDED

        gslgmp_linalg_symmtd_unpack (A, tau, evec, d, sd);

	gslgmp_vector_part(sd,0,N-1,tau); //CHECK IF NEEDED


     
//sfprintf(stderr,"Eigenval DEBUG1\n");



      // Make an initial pass through the tridiagonal decomposition
      //   to remove off-diagonal elements which are effectively zero 
      
      gmpchop_small_elements (N, d, sd);
      

// fprintf(stderr,"Eigenval DEBUG\n");
// 
// 	gmpPMat(A);
// 	gmpPMat(evec);
// 
// 	//gmpPVec(tau);
// 	gmpPVec(d);
// 	gmpPVec(sd);
// 
// fprintf(stderr,"END Eigenval DEBUG\n");

      // Progressively reduce the matrix until it is diagonal 
      
      b = N - 1;
      
      while (b > 0)
        {
          if ( mpfr_sgn(sd->data[b - 1]) == 0 || mpfr_nan_p (sd->data[b - 1])!=0)
            {
              b--;
              continue;
            }
          
          // Find the largest unreduced block (a,b) starting from b
          //   and working backwards 
          
          a = b - 1;
          
          while (a > 0)
            {
              if (mpfr_sgn(sd->data[a - 1]) == 0.0)
                {
                  break;
                }
              a--;
            }
          
          {
            int i;
            const int n_block = b - a + 1;

//             double *d_block = d + a;
//             double *sd_block = sd + a;
//		fprintf(stderr,"Eigenval DEBUG2\n");

		d_block=gslgmp_vector_calloc(n_block);
		sd_block=gslgmp_vector_calloc(n_block);

//		fprintf(stderr,"Eigenval DEBUG2.2\n");

		gslgmp_vector_part(d,a,n_block,d_block);
//		fprintf(stderr,"Eigenval DEBUG2.4\n");

		gslgmp_vector_partN(sd,sd->size,a,n_block,sd_block);

//		fprintf(stderr,"Eigenval DEBUG2.5\n");


//            double * const gc = w->gc;
//              double * const gs = w->gs;

		gc=gslgmp_vector_calloc(n_block);
		gs=gslgmp_vector_calloc(n_block);

            
            // apply QR reduction with implicit deflation to the
            //   unreduced block 
            
//		fprintf(stderr,"Eigenval DEBUG3\n");

            gmpqrstep (n_block, d_block, sd_block, gc, gs);
            


 	    	gslgmp_vector_reverse_part(d,a,n_block,d_block);
  	   	gslgmp_vector_reverse_partN(sd,N-1,a,n_block,sd_block);
//fprintf(stderr,"Eigenval DEBUG\n");

/* fprintf(stderr,"Eigenval DEBUG\n");
 
 	
 	gmpPVec(d);
 	gmpPVec(sd);
 	gmpPVec(d_block);
	gmpPVec(sd_block);
 	//gmpPVec(gc);
 	//gmpPVec(gs);

 fprintf(stderr,"END Eigenval DEBUG\n");

while(getchar()==0);
*/
            // Apply  Givens rotation Gij(c,s) to matrix Q,  Q <- Q G 
            
            for (i = 0; i < n_block - 1; i++)
              {
		gslgmp_vector_get(gc,i,c);
		gslgmp_vector_get(gs,i,s);

                //const double c = gc[i], s = gs[i];
                int k;
                
                for (k = 0; k < N; k++)
                  {
                    gslgmp_matrix_get (evec, k, a + i,qki);
                    gslgmp_matrix_get (evec, k, a + i + 1,qkj);

			mpfr_mul(tmp1,qki,c,GMP_RNDN);
			mpfr_mul(tmp2,qkj,s,GMP_RNDN);
			mpfr_sub(tmp3,tmp1,tmp2,GMP_RNDN);
                    gslgmp_matrix_set (evec, k, a + i, tmp3);
			
//                    gslgmp_matrix_set (evec, k, a + i, qki * c - qkj * s);
//                    gslgmp_matrix_set (evec, k, a + i + 1, qki * s + qkj * c);

			mpfr_mul(tmp1,qki,s,GMP_RNDN);
			mpfr_mul(tmp2,qkj,c,GMP_RNDN);
			mpfr_add(tmp3,tmp1,tmp2,GMP_RNDN);
                    gslgmp_matrix_set (evec, k, a + i + 1, tmp3);

                  }
              }
            


            // remove any small off-diagonal elements 
            
            gmpchop_small_elements (N, d, sd);
 
	  gslgmp_vector_free(d_block);
	  gslgmp_vector_free(sd_block);
	  gslgmp_vector_free(gc);
	  gslgmp_vector_free(gs);


         }



        }

      {
        //gsl_vector_view d_vec = gsl_vector_view_array (d, N);
        gslgmp_vector_memcpy (eval, d);
      }
      
	  gslgmp_vector_free(d);
	  gslgmp_vector_free(sd);
	  gslgmp_vector_free(tau);

	mpfr_clear( A00); 
	mpfr_clear( c); 
	mpfr_clear( tmp1); 
	mpfr_clear( tmp2); 
	mpfr_clear( tmp3); 
	mpfr_clear( s); 
	mpfr_clear( qki); 
	mpfr_clear( qkj); 
	mpfr_clear( zero); 
   
      return;
    }
}


void gslgmp_eigen_symm (gslgmp_matrix * A, gslgmp_vector * eval)
{

	fprintf(stderr, "Since this routine is not used currently\nI do not guarantee it functioning properly\nExiting.\n"); exit(-1);
	
#if (0)	
	
  if (A->size1 != A->size2)
    {
      fprintf(stderr, "matrix must be square to compute eigenvalues");
    }
  else if (eval->size != A->size1)
    {
      fprintf(stderr,"eigenvalue vector must match matrix size");
    }
   else
    {
 
     	const int N = A->size1;
      	int a, b;
     	mpfr_t A00,c,s,qki,qkj,zero,tmp1,tmp2,tmp3;
 	mpfr_init2(A00,PRECBITS);
 	mpfr_init2(c,PRECBITS);
 	mpfr_init2(s,PRECBITS);
 	mpfr_init2(qki,PRECBITS);
 	mpfr_init2(qkj,PRECBITS);
 	mpfr_init2(zero,PRECBITS);
	mpfr_init2(tmp1,PRECBITS);
	mpfr_init2(tmp2,PRECBITS);
	mpfr_init2(tmp3,PRECBITS);
	mpfr_set_d(zero,0.,GMP_RNDN);


      // handle special case 
      if (N == 1)
        {
          gslgmp_matrix_get (A, 0, 0,A00);
          gslgmp_vector_set (eval, 0, A00);
	  return;
         }

      // use sd as the temporary workspace for the decomposition when
      //   computing eigenvectors 

	gslgmp_vector* d;
	gslgmp_vector* sd;
	gslgmp_vector* tau;
	d=gslgmp_vector_calloc(N);
	sd=gslgmp_vector_calloc(N-1);
	tau=gslgmp_vector_calloc(N-1);

		gslgmp_vector *d_block;
		gslgmp_vector *sd_block;
		gslgmp_vector *gc;
		gslgmp_vector *gs;

	gc=gslgmp_vector_calloc(N);
	gs=gslgmp_vector_calloc(N);
		
      
       // gsl_vector_view d_vec = gsl_vector_view_array (d, N);
       // gsl_vector_view sd_vec = gsl_vector_view_array (sd, N - 1);
       // gsl_vector_view tau = gsl_vector_view_array (sd, N - 1);

	
        gslgmp_linalg_symmtd_decomp (A,tau);

	gslgmp_vector_reverse_part(sd,0,N-1,tau); //CHECK IF NEEDED

        gslgmp_linalg_symmtd_unpack_T (A, d, sd);

	gslgmp_vector_part(sd,0,N-1,tau); //CHECK IF NEEDED


     
fprintf(stderr,"Eigenval DEBUG1\n");



      // Make an initial pass through the tridiagonal decomposition
      //   to remove off-diagonal elements which are effectively zero 
      
      gmpchop_small_elements (N, d, sd);
      
fprintf(stderr,"Eigenval DEBUG2\n");

// fprintf(stderr,"Eigenval DEBUG\n");
// 
// 	gmpPMat(A);
// 	gmpPMat(evec);
// 
// 	//gmpPVec(tau);
// 	gmpPVec(d);
// 	gmpPVec(sd);
// 
// fprintf(stderr,"END Eigenval DEBUG\n");

      // Progressively reduce the matrix until it is diagonal 
      
      b = N - 1;
      
      while (b > 0)
        {
	 fprintf(stderr,"Eigenval DEBUG1 bloop\n");

	    
          if ( mpfr_sgn(sd->data[b - 1]) == 0 || mpfr_nan_p (sd->data[b - 1])!=0)
            {
              b--;
              continue;
            }
          
          // Find the largest unreduced block (a,b) starting from b
          //   and working backwards 
          
          a = b - 1;
          
          while (a > 0)
            {
             fprintf(stderr,"Eigenval DEBUG1 aloop\n");

		
              if (mpfr_sgn(sd->data[a - 1]) == 0.0)
                {
                  break;
                }
              a--;
            }
          
          {
            int i;
            const int n_block = b - a + 1;

//             double *d_block = d + a;
//             double *sd_block = sd + a;
//		fprintf(stderr,"Eigenval DEBUG2\n");

		d_block=gslgmp_vector_calloc(n_block);
		sd_block=gslgmp_vector_calloc(n_block);

//		fprintf(stderr,"Eigenval DEBUG2.2\n");
	fprintf(stderr,"Eigenval in loop 1 DEBUG\n");

		gslgmp_vector_part(d,a,n_block,d_block);
//		fprintf(stderr,"Eigenval DEBUG2.4\n");
	fprintf(stderr,"Eigenval in loop 2 DEBUG\n");

		gslgmp_vector_partN(sd,sd->size,a,n_block,sd_block);
	fprintf(stderr,"Eigenval in loop 3 DEBUG\n");

//		fprintf(stderr,"Eigenval DEBUG2.5\n");

           
            // apply QR reduction with implicit deflation to the
            //   unreduced block 
            
//		fprintf(stderr,"Eigenval DEBUG3\n");

            	gmpqrstep (n_block, d_block, sd_block, gc, gs);
            
	fprintf(stderr,"Eigenval in loop 4 DEBUG\n");


 	    	gslgmp_vector_reverse_part(d,a,n_block,d_block);
	fprintf(stderr,"Eigenval in loop 5 DEBUG\n");
  	   	gslgmp_vector_reverse_partN(sd,N-1,a,n_block,sd_block);
	fprintf(stderr,"Eigenval in loop 6 DEBUG\n");
           
            gmpchop_small_elements (N, d, sd);
 
	  gslgmp_vector_free(d_block);
	  gslgmp_vector_free(sd_block);


         }



        }

      {
 
	      gslgmp_vector_memcpy (eval, d);
      }
      
	  gslgmp_vector_free(gc);
	  gslgmp_vector_free(gs);
	  gslgmp_vector_free(d);
	  gslgmp_vector_free(sd);
	  gslgmp_vector_free(tau);

	mpfr_clear( A00); 
	mpfr_clear( c); 
	mpfr_clear( tmp1); 
	mpfr_clear( tmp2); 
	mpfr_clear( tmp3); 
	mpfr_clear( s); 
	mpfr_clear( qki); 
	mpfr_clear( qkj); 
	mpfr_clear( zero); 
   
      return;
    }
    
    #endif
}


void gmpEigenSYS_SYMM(gslgmp_matrix* in, gslgmp_matrix* Evec, gslgmp_vector* Eval)
{
	//LOWER DIAGONAL MATRIX IS BEING USED HERE!!

	if(in->size1!=in->size2||Evec->size1!=Evec->size2||Evec->size1!=in->size1||Eval->size!=in->size1){ fprintf(stderr,"EIGEN ERROR N!=M exiting\n"); exit(-1);}

	//TEST SYMMETRY OF THE MATRIX
	int nonsym=0;

	for(int i=0;i<in->size1;i++){	
	   for(int j=0;j<in->size2;j++){

	if( gslgmp_matrix_get_d(in,i,j)!=gslgmp_matrix_get_d(in,j,i)) nonsym=1;

	}}
	
	if(nonsym==1){ fprintf(stderr, "MATRIX IS NOT SYMMETRIC!! exiting\n"); exit(-1);}


	gslgmp_matrix* tmp;
	tmp=gslgmp_matrix_calloc(in->size1,in->size2);
	gslgmp_matrix_memcpy(tmp,in);

        gslgmp_eigen_symmv (tmp, Eval, Evec);


	gslgmp_matrix_free(tmp);

}

void gmpEigenVAL_SYMM(gslgmp_matrix* in, gslgmp_vector* Eval)
{
	//LOWER DIAGONAL MATRIX IS BEING USED HERE!!

	if(in->size1!=in->size2||Eval->size!=in->size1){ fprintf(stderr,"EIGEN ERROR N!=M exiting\n"); exit(-1);}

	//TEST SYMMETRY OF THE MATRIX
	int nonsym=0;

	for(int i=0;i<in->size1;i++){	
	   for(int j=0;j<in->size2;j++){

	if( gslgmp_matrix_get_d(in,i,j)!=gslgmp_matrix_get_d(in,j,i)) nonsym=1;

	}}
	
	if(nonsym==1){ fprintf(stderr, "MATRIX IS NOT SYMMETRIC!! exiting\n"); exit(-1);}


	gslgmp_matrix* tmp;
	tmp=gslgmp_matrix_calloc(in->size1,in->size2);
	gslgmp_matrix_memcpy(tmp,in);

        gslgmp_eigen_symm (tmp, Eval);


	gslgmp_matrix_free(tmp);

}

void gmpEigenSYS_SYMM_REVERSE(gslgmp_matrix* in, gslgmp_matrix* Evec, gslgmp_vector* Eval)
{
	//LOWER DIAGONAL MATRIX IS BEING USED HERE!!

	if(in->size1!=in->size2||Evec->size1!=Evec->size2||Evec->size1!=in->size1||Eval->size!=in->size1){ fprintf(stderr,"EIGEN ERROR N!=M exiting\n"); exit(-1);}

	//TEST SYMMETRY OF THE MATRIX
	int nonsym=0;

	for(int i=0;i<in->size1;i++){	
	   for(int j=0;j<in->size2;j++){

	if( gslgmp_matrix_get_d(in,i,j)!=gslgmp_matrix_get_d(in,j,i)) nonsym=1;

	}}
	
	if(nonsym==1){ fprintf(stderr, "MATRIX IS NOT SYMMETRIC!! exiting\n"); exit(-1);}

	gslgmp_matrix* tmp;
	tmp=gslgmp_matrix_calloc(in->size1,in->size2);
	gslgmp_matrix* tmp2;
	tmp2=gslgmp_matrix_calloc(in->size1,in->size2);
	gslgmp_vector* tmp3;
	tmp3=gslgmp_vector_calloc(in->size1);

	for(int i=0;i<in->size1;i++){
	   for(int j=0;j<in->size2;j++){		
		mpfr_set( tmp->data[(in->size1-i-1)*(in->size2)+j], in->data[i*(in->size2)+j], GMP_RNDN);
	}}

 
        gslgmp_eigen_symmv (tmp, tmp3, tmp2);


	for(int i=0;i<in->size1;i++){
	   for(int j=0;j<in->size2;j++){		
		mpfr_set(Evec->data[(in->size1-i-1)*(in->size2)+j],tmp2->data[i*(in->size2)+j],GMP_RNDN);
	}}

	gslgmp_matrix_free(tmp);
	gslgmp_matrix_free(tmp2);
	gslgmp_vector_free(tmp3);
}

//**********************************************************************************
//  Testing routines for Errors!!
//**********************************************************************************

void AgleichB(gslgmp_matrix* in1, gslgmp_matrix* in2)
{
	if(in1->size1!=in2->size1||in1->size2!=in2->size2) {fprintf(stderr, "MATRICES DO NOT MATCH\n");exit(-1); }

	mpfr_t errtot,ptmp1,ptmp2,ptmp3;
	mpfr_init2(errtot,PRECBITS);
	mpfr_init2(ptmp1,PRECBITS);
	mpfr_init2(ptmp2,PRECBITS);
	mpfr_init2(ptmp3,PRECBITS);
	mpfr_set_d(errtot,0.,GMP_RNDN);

	for(int i=0;i<in1->size1;i++){
	  for(int j=0;j<in1->size2;j++){

		gslgmp_matrix_get(in1,i,j,ptmp1);
		gslgmp_matrix_get(in2,i,j,ptmp2);
		
		mpfr_sub(ptmp3,ptmp1,ptmp2,GMP_RNDN);
		mpfr_abs(ptmp3,ptmp3,GMP_RNDN);

		mpfr_set(ptmp1,errtot,GMP_RNDN);
		mpfr_add(errtot,ptmp1,ptmp3,GMP_RNDN);
	
	}}
	
#if (VERBOSE==1)
	fprintf(stderr, "Error between matrices is: %g\n\n",mpfr_get_d(errtot,GMP_RNDN));
#endif
	mpfr_clear(errtot);
	mpfr_clear(ptmp1);
	mpfr_clear(ptmp2);
	mpfr_clear(ptmp3);


}

void AgleichB(gslgmp_matrix* in1, gslgmp_matrix* in2, double* ERROR)
{
	if(in1->size1!=in2->size1||in1->size2!=in2->size2) {fprintf(stderr, "MATRICES DO NOT MATCH\n");exit(-1); }

	mpfr_t errtot,ptmp1,ptmp2,ptmp3;
	mpfr_init2(errtot,PRECBITS);
	mpfr_init2(ptmp1,PRECBITS);
	mpfr_init2(ptmp2,PRECBITS);
	mpfr_init2(ptmp3,PRECBITS);
	mpfr_set_d(errtot,0.,GMP_RNDN);

	for(int i=0;i<in1->size1;i++){
	  for(int j=0;j<in1->size2;j++){

		gslgmp_matrix_get(in1,i,j,ptmp1);
		gslgmp_matrix_get(in2,i,j,ptmp2);
		
		mpfr_sub(ptmp3,ptmp1,ptmp2,GMP_RNDN);
		mpfr_abs(ptmp3,ptmp3,GMP_RNDN);

		mpfr_set(ptmp1,errtot,GMP_RNDN);
		mpfr_add(errtot,ptmp1,ptmp3,GMP_RNDN);
	
	}}
	
#if (VERBOSE==1)
	fprintf(stderr, "Error between matrices is: %g\n\n",mpfr_get_d(errtot,GMP_RNDN));
#endif


	*ERROR=mpfr_get_d(errtot,GMP_RNDN);

	mpfr_clear(errtot);
	mpfr_clear(ptmp1);
	mpfr_clear(ptmp2);
	mpfr_clear(ptmp3);


}

void AgleichBt(gslgmp_matrix* in1, gslgmp_matrix* in2)
{
	if(in1->size1!=in2->size2||in1->size2!=in2->size1) {fprintf(stderr, "MATRICES DO NOT MATCH\n");exit(-1); }

	mpfr_t errtot,ptmp1,ptmp2,ptmp3;
	mpfr_init2(errtot,PRECBITS);
	mpfr_init2(ptmp1,PRECBITS);
	mpfr_init2(ptmp2,PRECBITS);
	mpfr_init2(ptmp3,PRECBITS);
	mpfr_set_ui(errtot,0,GMP_RNDN);

	for(int i=0;i<in1->size1;i++){
	  for(int j=0;j<in1->size2;j++){

		gslgmp_matrix_get(in1,i,j,ptmp1);
		gslgmp_matrix_get(in2,j,i,ptmp2);
		
		mpfr_sub(ptmp3,ptmp1,ptmp2,GMP_RNDN);
		mpfr_abs(ptmp3,ptmp3,GMP_RNDN);

		mpfr_set(ptmp1,errtot,GMP_RNDN);
		mpfr_add(errtot,ptmp1,ptmp3,GMP_RNDN);
	
	}}
	
	fprintf(stderr, "Error between matrices is: %g\n\n",mpfr_get_d(errtot,GMP_RNDN));
	mpfr_clear(errtot);
	mpfr_clear(ptmp1);
	mpfr_clear(ptmp2);
	mpfr_clear(ptmp3);

}

void Agleich1(gslgmp_matrix* in1)
{
	if(in1->size1!=in1->size2) {fprintf(stderr, "MATRIX NOT SQUARE\n");exit(-1); }

	double errtot=0.;

	for(int i=0;i<in1->size1;i++){
	  for(int j=0;j<in1->size2;j++){

	    if(i==j){
	    errtot+=fabs(gslgmp_matrix_get_d(in1,i,j)-1.);
	    }else{
	    errtot+=fabs(gslgmp_matrix_get_d(in1,i,j));
	    }

	}}
	
	fprintf(stderr, "Deviation from unit matrix is: %g\n\n",errtot);

}


void ugleichv(gslgmp_vector* in1, gslgmp_vector* in2)
{
	if(in1->size!=in2->size) {fprintf(stderr, "Vectors DO NOT MATCH\n");exit(-1); }

	double errtot=0.;

	for(int i=0;i<in1->size;i++){
	 

	    errtot+=fabs(gslgmp_vector_get_d(in1,i)-gslgmp_vector_get_d(in2,i));
	
	}
	
	fprintf(stderr, "Error between vectors is: %g\n\n",errtot);

}

void SVD_test(gslgmp_matrix* original,gslgmp_matrix* U,gslgmp_matrix* V,gslgmp_vector* Xi)
{
	gslgmp_matrix *TestMat;
	TestMat=gslgmp_matrix_calloc(original->size1,original->size2);

	gmp_multmvmt(U,Xi,V,TestMat);

	fprintf(stderr,"SVD Test:\n");
	AgleichB(TestMat,original);

	gslgmp_matrix_free(TestMat);

}

void SVDTrans_test(gslgmp_matrix* original,gslgmp_matrix* U,gslgmp_matrix* V,gslgmp_vector* Xi)
{
	gslgmp_matrix *TestMat;
	TestMat=gslgmp_matrix_calloc(original->size2,original->size1);

	gmp_multmvmt(U,Xi,V,TestMat);

	fprintf(stderr,"SVD Test:\n");
	AgleichBt(TestMat,original);

	gslgmp_matrix_free(TestMat);

}

void EiV_test(gslgmp_matrix* original,gslgmp_matrix* Evec,gslgmp_vector* Eval)
{
	gslgmp_matrix *TestMat;
	TestMat=gslgmp_matrix_calloc(original->size1,original->size2);

	gslgmp_matrix *TestVal;
	TestVal=gslgmp_matrix_calloc(original->size1,original->size2);

	gmp_VMakeM(Eval,TestVal);

	gmp_mult3m(Evec,0,TestVal,0,Evec,1,TestMat);

	fprintf(stderr,"Eigen decomposition Test:\n");
	AgleichB(TestMat,original);

	gslgmp_matrix_free(TestVal);
	gslgmp_matrix_free(TestMat);

}

void EiV_test(gslgmp_matrix* original,gslgmp_matrix* Evec,gslgmp_vector* Eval, double* ERR)
{
	gslgmp_matrix *TestMat;
	TestMat=gslgmp_matrix_calloc(original->size1,original->size2);

	gslgmp_matrix *TestVal;
	TestVal=gslgmp_matrix_calloc(original->size1,original->size2);

	gmp_VMakeM(Eval,TestVal);

	gmp_mult3m(Evec,0,TestVal,0,Evec,1,TestMat);

#if (VERBOSE==1)
	fprintf(stderr,"Eigen decomposition Test:\n");
#endif
	AgleichB(TestMat,original,ERR);

	gslgmp_matrix_free(TestVal);
	gslgmp_matrix_free(TestMat);

}

void INV_test(gslgmp_matrix* original,gslgmp_matrix* invert)
{
	gslgmp_matrix *TestMat;
	TestMat=gslgmp_matrix_calloc(original->size1,original->size2);

	gmp_mult2m(original,0,invert,0,TestMat);

	fprintf(stderr,"Inversion Test:\n");
	Agleich1(TestMat);

	gslgmp_matrix_free(TestMat);
}

void ORT_test(gslgmp_matrix* original)
{
	if(original->size1!=original->size2){fprintf(stderr,"Matrix is not quadratic\n");exit(-1);}

	mpfr_t tmp;
	mpfr_init2(tmp,PRECBITS);

	gslgmp_matrix *TestMat;
	TestMat=gslgmp_matrix_calloc(original->size1,original->size2);

	gmp_mult2m(original,0,original,1,TestMat);

	fprintf(stderr,"Orthogonality Test:\n");
	Agleich1(TestMat);

}

void SYM_test(gslgmp_matrix* original)
{
	if(original->size1!=original->size2){fprintf(stderr,"Matrix is not quadratic\n");exit(-1);}

	fprintf(stderr,"Symmetry Test:\n");
	AgleichBt(original,original);

}
